"""Configuration options for importing data into Neuracore.

This module defines options of the configuration for specifying the format
of the input data.

"""

from enum import Enum

from pydantic import BaseModel, Field, model_validator


class DatasetTypeConfig(str, Enum):
    """Enumeration of supported dataset types."""

    RLDS = "RLDS"
    LEROBOT = "LEROBOT"
    TFDS = "TFDS"


class OutputDatasetConfig(BaseModel):
    """Output dataset configuration."""

    name: str
    tags: list[str] = Field(default_factory=list)
    description: str = ""


class RobotConfig(BaseModel):
    """Robot configuration."""

    name: str
    urdf_path: str | None = None
    mjcf_path: str | None = None
    overwrite_existing: bool = False


class VisualJointTypeConfig(str, Enum):
    """Types of visual joint positions."""

    GRIPPER = "GRIPPER"
    CUSTOM = "CUSTOM"


class JointPositionTypeConfig(str, Enum):
    """Types of joint positions."""

    END_EFFECTOR = "END_EFFECTOR"
    CUSTOM = "CUSTOM"


class RotationConfig(str, Enum):
    """Types of rotations."""

    QUATERNION = "QUATERNION"
    MATRIX = "MATRIX"
    EULER = "EULER"
    AXIS_ANGLE = "AXIS_ANGLE"


class AngleConfig(str, Enum):
    """Types of angles."""

    DEGREES = "DEGREES"
    RADIANS = "RADIANS"


class PoseConfig(str, Enum):
    """Types of poses."""

    MATRIX = "MATRIX"
    POSITION_ORIENTATION = "POSITION_ORIENTATION"


class QuaternionOrderConfig(str, Enum):
    """Order of quaternion."""

    XYZW = "XYZW"
    WXYZ = "WXYZ"


class EulerOrderConfig(str, Enum):
    """Order of euler angles."""

    XYZ = "XYZ"
    ZYX = "ZYX"
    YXZ = "YXZ"
    ZXY = "ZXY"
    YZX = "YZX"
    XZY = "XZY"


class IndexRangeConfig(BaseModel):
    """Configuration for index range of data extraction."""

    start: int
    end: int

    @model_validator(mode="after")
    def validate_index_range(self) -> "IndexRangeConfig":
        """Validate that index range is valid."""
        if self.start > self.end:
            raise ValueError("Index range start must be less than end")
        return self


class NormalizeConfig(BaseModel):
    """Configuration for normalizing data."""

    min: float = 0.0
    max: float = 1.0


class ImageConventionConfig(str, Enum):
    """Convention of image channels."""

    CHANNELS_LAST = "CHANNELS_LAST"
    CHANNELS_FIRST = "CHANNELS_FIRST"


class ImageChannelOrderConfig(str, Enum):
    """Order of image channels."""

    RGB = "RGB"
    BGR = "BGR"


class LanguageConfig(str, Enum):
    """Types of languages."""

    STRING = "STRING"
    BYTES = "BYTES"


class TorqueUnitsConfig(str, Enum):
    """Types of torque units."""

    NM = "NM"
    NCM = "NCM"


class DistanceUnitsConfig(str, Enum):
    """Types of distance units."""

    M = "M"
    MM = "MM"


class OrientationConfig(BaseModel):
    """Configuration for orientation of poses."""

    type: RotationConfig = RotationConfig.QUATERNION
    quaternion_order: QuaternionOrderConfig = QuaternionOrderConfig.XYZW
    euler_order: EulerOrderConfig = EulerOrderConfig.XYZ
    angle_units: AngleConfig = AngleConfig.RADIANS



"""Unit tests for config.py module."""

import json

import pytest
import yaml

from neuracore_types.importer.config import (
    AngleConfig,
    DistanceUnitsConfig,
    EulerOrderConfig,
    ImageChannelOrderConfig,
    ImageConventionConfig,
    IndexRangeConfig,
    LanguageConfig,
    NormalizeConfig,
    OrientationConfig,
    OutputDatasetConfig,
    PoseConfig,
    QuaternionOrderConfig,
    RobotConfig,
    RotationConfig,
    TorqueUnitsConfig,
)
from neuracore_types.importer.data_config import DataFormat, MappingItem
from neuracore_types.importer.transform import (
    CastToNumpyDtype,
    Clip,
    ImageFormat,
    NumpyToScalar,
)
from neuracore_types.nc_data import DatasetImportConfig, DataType
from neuracore_types.nc_data.camera_data import RGBCameraDataImportConfig
from neuracore_types.nc_data.joint_data import JointPositionsDataImportConfig


class TestOutputDatasetConfig:
    """Tests for OutputDatasetConfig class."""

    def test_output_dataset_config_basic(self):
        """Test basic OutputDatasetConfig creation."""
        dataset = OutputDatasetConfig(
            name="test_dataset", description="Test description"
        )
        assert dataset.name == "test_dataset"
        assert dataset.description == "Test description"
        assert dataset.tags == []

    def test_output_dataset_config_with_tags(self):
        """Test OutputDatasetConfig with tags."""
        dataset = OutputDatasetConfig(
            name="test_dataset", tags=["tag1", "tag2"], description="Test"
        )
        assert dataset.tags == ["tag1", "tag2"]


class TestRobotConfig:
    """Tests for RobotConfig class."""

    def test_robot_config_basic(self):
        """Test basic RobotConfig creation."""
        robot = RobotConfig(name="test_robot")
        assert robot.name == "test_robot"
        assert robot.urdf_path is None
        assert robot.mjcf_path is None
        assert robot.overwrite_existing is False

    def test_robot_config_with_paths(self):
        """Test RobotConfig with URDF and MJCF paths."""
        robot = RobotConfig(
            name="test_robot",
            urdf_path="/path/to/urdf",
            mjcf_path="/path/to/mjcf",
            overwrite_existing=True,
        )
        assert robot.urdf_path == "/path/to/urdf"
        assert robot.mjcf_path == "/path/to/mjcf"
        assert robot.overwrite_existing is True


class TestIndexRangeConfig:
    """Tests for IndexRangeConfig class."""

    def test_index_range_valid(self):
        """Test valid IndexRangeConfig."""
        index_range = IndexRangeConfig(start=0, end=10)
        assert index_range.start == 0
        assert index_range.end == 10

    def test_index_range_invalid(self):
        """Test invalid IndexRangeConfig where start > end."""
        with pytest.raises(ValueError, match="Index range start must be less than end"):
            IndexRangeConfig(start=10, end=0)


class TestNormalizeConfig:
    """Tests for NormalizeConfig class."""

    def test_normalize_config_defaults(self):
        """Test NormalizeConfig with default values."""
        config = NormalizeConfig()
        assert config.min == 0.0
        assert config.max == 1.0

    def test_normalize_config_custom(self):
        """Test NormalizeConfig with custom values."""
        config = NormalizeConfig(min=-1.0, max=1.0)
        assert config.min == -1.0
        assert config.max == 1.0


class TestMappingItem:
    """Tests for MappingItem class."""

    def test_mapping_item_basic(self):
        """Test basic MappingItem creation."""
        item = MappingItem(name="joint_0")
        assert item.name == "joint_0"
        assert item.source_name is None
        assert item.index is None
        assert item.index_range is None
        assert item.offset == 0.0
        assert item.inverted is False

    def test_mapping_item_with_index(self):
        """Test MappingItem with index."""
        item = MappingItem(name="joint_0", index=5)
        assert item.index == 5

    def test_mapping_item_with_index_range(self):
        """Test MappingItem with index_range."""
        index_range = IndexRangeConfig(start=0, end=3)
        item = MappingItem(name="pose", index_range=index_range)
        assert item.index_range == index_range

    def test_mapping_item_index_and_range_conflict(self):
        """Test MappingItem validation when both index and index_range are provided."""
        index_range = IndexRangeConfig(start=0, end=3)
        with pytest.raises(
            ValueError, match="index and index_range cannot be provided together"
        ):
            MappingItem(name="pose", index=5, index_range=index_range)


class TestOrientationConfig:
    """Tests for OrientationConfig class."""

    def test_orientation_config_defaults(self):
        """Test OrientationConfig with default values."""
        config = OrientationConfig()
        assert config.type == RotationConfig.QUATERNION
        assert config.quaternion_order == QuaternionOrderConfig.XYZW
        assert config.euler_order == EulerOrderConfig.XYZ
        assert config.angle_units == AngleConfig.RADIANS

    def test_orientation_config_euler(self):
        """Test OrientationConfig with euler type."""
        config = OrientationConfig(
            type=RotationConfig.EULER,
            euler_order=EulerOrderConfig.ZYX,
            angle_units=AngleConfig.DEGREES,
        )
        assert config.type == RotationConfig.EULER
        assert config.euler_order == EulerOrderConfig.ZYX
        assert config.angle_units == AngleConfig.DEGREES


class TestDataFormat:
    """Tests for DataFormat class."""

    def test_data_format_defaults(self):
        """Test DataFormat with default values."""
        fmt = DataFormat()
        assert fmt.image_convention == ImageConventionConfig.CHANNELS_LAST
        assert fmt.order_of_channels == ImageChannelOrderConfig.RGB
        assert fmt.normalized_pixel_values is False
        assert fmt.angle_units == AngleConfig.RADIANS
        assert fmt.torque_units == TorqueUnitsConfig.NM
        assert fmt.distance_units == DistanceUnitsConfig.M
        assert fmt.pose_type == PoseConfig.MATRIX
        assert fmt.orientation is None
        assert fmt.language_type == LanguageConfig.STRING
        assert fmt.normalize is None

    def test_data_format_custom(self):
        """Test DataFormat with custom values."""
        normalize = NormalizeConfig(min=0.0, max=1.0)
        orientation = OrientationConfig(type=RotationConfig.EULER)
        fmt = DataFormat(
            image_convention=ImageConventionConfig.CHANNELS_FIRST,
            order_of_channels=ImageChannelOrderConfig.BGR,
            normalized_pixel_values=True,
            angle_units=AngleConfig.DEGREES,
            torque_units=TorqueUnitsConfig.NCM,
            distance_units=DistanceUnitsConfig.MM,
            pose_type=PoseConfig.POSITION_ORIENTATION,
            orientation=orientation,
            language_type=LanguageConfig.BYTES,
            normalize=normalize,
        )
        assert fmt.image_convention == ImageConventionConfig.CHANNELS_FIRST
        assert fmt.order_of_channels == ImageChannelOrderConfig.BGR
        assert fmt.normalized_pixel_values is True
        assert fmt.angle_units == AngleConfig.DEGREES
        assert fmt.torque_units == TorqueUnitsConfig.NCM
        assert fmt.distance_units == DistanceUnitsConfig.MM
        assert fmt.pose_type == PoseConfig.POSITION_ORIENTATION
        assert fmt.orientation == orientation
        assert fmt.language_type == LanguageConfig.BYTES
        assert fmt.normalize == normalize


class TestDatasetConfig:
    """Tests for DatasetConfig class."""

    def test_dataset_config_basic(self):
        """Test basic DatasetConfig creation."""
        output_dataset = OutputDatasetConfig(name="output_dataset")
        robot = RobotConfig(name="test_robot")
        config = DatasetImportConfig(
            input_dataset_name="input_dataset",
            output_dataset=output_dataset,
            robot=robot,
        )
        assert config.input_dataset_name == "input_dataset"
        assert config.output_dataset == output_dataset
        assert config.robot == robot
        assert config.frequency is None
        assert config.data_import_config == {}

    def test_dataset_config_from_file_yaml(self, tmp_path):
        """Test DatasetConfig.from_file with YAML file"""
        config_path = tmp_path / "config.yaml"
        config_data = {
            "input_dataset_name": "input_dataset",
            "output_dataset": {"name": "output_dataset"},
            "robot": {"name": "test_robot"},
            "data_import_config": {
                "RGB_IMAGES": {
                    "source": "camera0",
                    "format": {
                        "image_convention": "CHANNELS_FIRST",
                        "order_of_channels": "RGB",
                        "normalized_pixel_values": False,
                    },
                    "mapping": [{
                        "name": "image",
                        "source_name": "camera0",
                    }],
                },
                "JOINT_POSITIONS": {
                    "source": "joint_sensor0",
                    "format": {
                        "angle_units": "RADIANS",
                    },
                    "mapping": [{
                        "name": "joint_0",
                        "index": 0,
                        "offset": 0.0,
                        "inverted": False,
                    }],
                },
            },
        }
        with config_path.open("w") as f:
            yaml.dump(config_data, f)

        config = DatasetImportConfig.from_file(config_path)
        assert config.input_dataset_name == "input_dataset"
        assert config.output_dataset.name == "output_dataset"
        assert config.robot.name == "test_robot"
        # Verify data_import_config contains the sample types
        assert DataType.RGB_IMAGES in config.data_import_config
        assert config.data_import_config[DataType.RGB_IMAGES].source == "camera0"
        assert DataType.JOINT_POSITIONS in config.data_import_config
        assert (
            config.data_import_config[DataType.JOINT_POSITIONS].source
            == "joint_sensor0"
        )
        # Check transforms
        rgb_transforms = (
            config.data_import_config[DataType.RGB_IMAGES]
            .mapping[0]
            .transforms.transforms
        )
        joint_transforms = (
            config.data_import_config[DataType.JOINT_POSITIONS]
            .mapping[0]
            .transforms.transforms
        )

        assert len(rgb_transforms) == 3
        assert len(joint_transforms) == 1

        assert isinstance(rgb_transforms[0], Clip)
        assert isinstance(rgb_transforms[1], CastToNumpyDtype)
        assert isinstance(rgb_transforms[2], ImageFormat)
        assert isinstance(joint_transforms[0], NumpyToScalar)

    def test_dataset_config_from_file_json(self, tmp_path):
        """Test DatasetConfig.from_file with JSON file."""
        config_path = tmp_path / "config.json"
        config_data = {
            "input_dataset_name": "input_dataset",
            "output_dataset": {"name": "output_dataset"},
            "robot": {"name": "test_robot"},
            "data_import_config": {
                "RGB_IMAGES": {
                    "source": "camera0",
                    "format": {
                        "image_convention": "CHANNELS_FIRST",
                        "order_of_channels": "RGB",
                        "normalized_pixel_values": False,
                    },
                    "mapping": [{
                        "name": "image",
                        "source_name": "camera0",
                    }],
                },
                "JOINT_POSITIONS": {
                    "source": "joint_sensor0",
                    "format": {
                        "angle_units": "RADIANS",
                    },
                    "mapping": [{
                        "name": "joint_0",
                        "index": 0,
                        "offset": 0.0,
                        "inverted": False,
                    }],
                },
            },
        }
        with config_path.open("w") as f:
            json.dump(config_data, f)

        config = DatasetImportConfig.from_file(config_path)
        assert config.input_dataset_name == "input_dataset"
        assert config.output_dataset.name == "output_dataset"
        assert config.robot.name == "test_robot"
        # Verify data_import_config contains the sample types
        assert DataType.RGB_IMAGES in config.data_import_config
        assert config.data_import_config[DataType.RGB_IMAGES].source == "camera0"
        assert DataType.JOINT_POSITIONS in config.data_import_config
        assert (
            config.data_import_config[DataType.JOINT_POSITIONS].source
            == "joint_sensor0"
        )
        # Check transforms
        rgb_transforms = (
            config.data_import_config[DataType.RGB_IMAGES]
            .mapping[0]
            .transforms.transforms
        )
        joint_transforms = (
            config.data_import_config[DataType.JOINT_POSITIONS]
            .mapping[0]
            .transforms.transforms
        )

        assert len(rgb_transforms) == 3
        assert len(joint_transforms) == 1

        assert isinstance(rgb_transforms[0], Clip)
        assert isinstance(rgb_transforms[1], CastToNumpyDtype)
        assert isinstance(rgb_transforms[2], ImageFormat)
        assert isinstance(joint_transforms[0], NumpyToScalar)

    def test_dataset_config_from_file_not_found(self, tmp_path):
        """Test DatasetConfig.from_file with non-existent file."""
        config_path = tmp_path / "missing_config.yaml"
        with pytest.raises(FileNotFoundError):
            DatasetImportConfig.from_file(config_path)

    def test_dataset_config_from_file_invalid_format(self, tmp_path):
        """Test DatasetConfig.from_file with invalid file format."""
        config_path = tmp_path / "config.txt"
        config_path.write_text("some content")
        with pytest.raises(RuntimeError, match="Unsupported config format"):
            DatasetImportConfig.from_file(config_path)

    def test_dataset_config_from_file_invalid_yaml(self, tmp_path):
        """Test DatasetConfig.from_file with invalid YAML."""
        config_path = tmp_path / "config.yaml"
        config_path.write_text("invalid: yaml: content: [")
        with pytest.raises(RuntimeError):
            DatasetImportConfig.from_file(config_path)

    def test_dataset_config_from_file_invalid_json(self, tmp_path):
        """Test DatasetConfig.from_file with invalid JSON."""
        config_path = tmp_path / "config.json"
        config_path.write_text("{invalid json}")
        with pytest.raises(RuntimeError):
            DatasetImportConfig.from_file(config_path)

    def test_dataset_config_with_data_points(self):
        """Test DatasetConfig with data points."""
        output_dataset = OutputDatasetConfig(name="output_dataset")
        robot = RobotConfig(name="test_robot")
        rgb_point = RGBCameraDataImportConfig(source="camera")
        joint_point = JointPositionsDataImportConfig(source="joints")
        data_import_config = {
            DataType.RGB_IMAGES: rgb_point,
            DataType.JOINT_POSITIONS: joint_point,
        }

        config = DatasetImportConfig(
            input_dataset_name="input_dataset",
            output_dataset=output_dataset,
            robot=robot,
            data_import_config=data_import_config,
        )
        assert config.data_import_config[DataType.RGB_IMAGES] == rgb_point
        assert config.data_import_config[DataType.JOINT_POSITIONS] == joint_point



"""Data models for natural language data."""

from typing import Literal

from pydantic import ConfigDict, Field

from neuracore_types.importer.config import LanguageConfig
from neuracore_types.importer.transform import (
    DataTransform,
    DataTransformSequence,
    LanguageFromBytes,
)
from neuracore_types.nc_data.nc_data import (
    DataItemStats,
    NCData,
    NCDataImportConfig,
    NCDataStats,
)
from neuracore_types.utils.pydantic_to_ts import (
    REQUIRED_WITH_DEFAULT_FLAG,
    fix_required_with_defaults,
)


class LanguageDataStats(NCDataStats):
    """Statistics for LanguageData."""

    type: Literal["LanguageDataStats"] = Field(
        default="LanguageDataStats", json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    text: DataItemStats

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)


class LanguageDataImportConfig(NCDataImportConfig):
    """Import configuration for LanguageData."""

    def _populate_transforms(self) -> None:
        """Populate transforms based on configuration."""
        transform_list: list[DataTransform] = []
        # Add LanguageFromBytes transform if configured for bytes
        if self.format.language_type == LanguageConfig.BYTES:
            transform_list.append(LanguageFromBytes())
        for item in self.mapping:
            item.transforms = DataTransformSequence(transforms=transform_list)


class LanguageData(NCData):
    """Natural language instruction or description data.

    Contains text-based information such as task descriptions, voice commands,
    or other linguistic data associated with robot demonstrations.
    """

    type: Literal["LanguageData"] = Field(
        default="LanguageData", json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    text: str

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)

    @classmethod
    def sample(cls) -> "LanguageData":
        """Sample an example LanguageData instance.

        Returns:
            LanguageData: Sampled LanguageData instance
        """
        return cls(text="Sample instruction.")

    def calculate_statistics(self) -> LanguageDataStats:
        """Calculate the statistics for this data type."""
        return LanguageDataStats(text=DataItemStats())



"""Unit tests for transform.py module."""

import numpy as np
import pytest
from scipy.spatial.transform import Rotation as R

from neuracore_types.importer.config import (
    AngleConfig,
    EulerOrderConfig,
    ImageChannelOrderConfig,
    ImageConventionConfig,
    PoseConfig,
    QuaternionOrderConfig,
    RotationConfig,
)
from neuracore_types.importer.transform import (
    CastToNumpyDtype,
    Clip,
    DataTransformSequence,
    DegreesToRadians,
    FlipSign,
    ImageChannelOrder,
    ImageFormat,
    LanguageFromBytes,
    NanToNum,
    Normalize,
    NumpyToScalar,
    Offset,
    Pose,
    Rotation,
    Scale,
    Unnormalize,
)


class TestDataTransformSequence:
    """Tests for DataTransformSequence class."""

    def test_data_transform_sequence_empty(self):
        """Test DataTransformSequence with empty transforms."""
        sequence = DataTransformSequence(transforms=[])
        data = np.array([1, 2, 3])
        result = sequence(data)
        np.testing.assert_array_equal(result, data)

    def test_data_transform_sequence_multiple_transforms(self):
        """Test DataTransformSequence with multiple transforms."""
        sequence = DataTransformSequence(
            transforms=[Scale(factor=2.0), Offset(value=1.0)]
        )
        data = np.array([1.0, 2.0, 3.0])
        result = sequence(data)
        expected = np.array([3.0, 5.0, 7.0])  # (x * 2) + 1
        np.testing.assert_array_equal(result, expected)


class TestRotation:
    """Tests for Rotation transform."""

    def test_rotation_quaternion_xyzw(self):
        """Test Rotation with quaternion xyzw (no change)."""
        transform = Rotation(
            rotation_type=RotationConfig.QUATERNION,
            angle_type=AngleConfig.RADIANS,
            seq=QuaternionOrderConfig.XYZW,
        )
        quat = np.array([0.5, 0.5, 0.5, 0.5])
        result = transform(quat)
        np.testing.assert_array_almost_equal(result, quat)

    def test_rotation_quaternion_wxyz(self):
        """Test Rotation with quaternion wxyz."""
        transform = Rotation(
            rotation_type=RotationConfig.QUATERNION,
            angle_type=AngleConfig.RADIANS,
            seq=QuaternionOrderConfig.WXYZ,
        )
        quat_wxyz = np.array([1.0, 0.0, 0.0, 0.0])  # w, x, y, z
        result = transform(quat_wxyz)
        expected = np.array([0.0, 0.0, 0.0, 1.0])
        np.testing.assert_array_almost_equal(result, expected)

    def test_rotation_matrix(self):
        """Test Rotation with rotation matrix."""
        transform = Rotation(
            rotation_type=RotationConfig.MATRIX, angle_type=AngleConfig.RADIANS
        )
        # Identity rotation matrix
        rot_matrix = np.eye(3)
        result = transform(rot_matrix)
        # Identity matrix should give quaternion [0, 0, 0, 1] (xyzw)
        expected = np.array([0.0, 0.0, 0.0, 1.0])
        np.testing.assert_array_almost_equal(result, expected)

    def test_rotation_euler_radians(self):
        """Test Rotation with euler angles in radians."""
        transform = Rotation(
            rotation_type=RotationConfig.EULER,
            angle_type=AngleConfig.RADIANS,
            seq=EulerOrderConfig.XYZ.value,
        )
        euler = np.array([0.0, 0.0, 0.0])
        result = transform(euler)
        # Zero euler should give quaternion [0, 0, 0, 1]
        expected = np.array([0.0, 0.0, 0.0, 1.0])
        np.testing.assert_array_almost_equal(result, expected)

    def test_rotation_euler_degrees(self):
        """Test Rotation with euler angles in degrees."""
        transform = Rotation(
            rotation_type=RotationConfig.EULER,
            angle_type=AngleConfig.DEGREES,
            seq=EulerOrderConfig.XYZ,
        )
        euler = np.array([90.0, 0.0, 0.0])
        result = transform(euler)
        # 90 degree rotation around x-axis
        expected = R.from_euler("xyz", [90.0, 0.0, 0.0], degrees=True).as_quat()
        np.testing.assert_array_almost_equal(result, expected)

    def test_rotation_axis_angle_radians(self):
        """Test Rotation with axis-angle in radians."""
        transform = Rotation(
            rotation_type=RotationConfig.AXIS_ANGLE, angle_type=AngleConfig.RADIANS
        )
        axis_angle = np.array([0.0, 0.0, 0.0])
        result = transform(axis_angle)
        expected = np.array([0.0, 0.0, 0.0, 1.0])
        np.testing.assert_array_almost_equal(result, expected)

    def test_rotation_axis_angle_degrees(self):
        """Test Rotation with axis-angle in degrees."""
        transform = Rotation(
            rotation_type=RotationConfig.AXIS_ANGLE, angle_type=AngleConfig.DEGREES
        )
        # 90 degree rotation around z-axis
        axis_angle_deg = np.array([0.0, 0.0, 90.0])
        result = transform(axis_angle_deg)
        expected = R.from_rotvec([0.0, 0.0, np.pi / 2]).as_quat()
        np.testing.assert_array_almost_equal(result, expected)

    def test_rotation_unsupported_type(self):
        """Test Rotation with a fake rotation type to hit the error path."""
        with pytest.raises(Exception):  # Pydantic will raise validation error
            Rotation(rotation_type="unsupported")


class TestPose:
    """Tests for Pose transform."""

    def test_pose_matrix(self):
        """Test Pose with matrix format."""
        transform = Pose(pose_type=PoseConfig.MATRIX)
        # 4x4 transformation matrix (identity)
        pose_matrix = np.eye(4)
        result = transform(pose_matrix)
        # Should extract position (last column) and rotation (3x3 matrix)
        expected_pos = pose_matrix[:3, 3]
        expected_rot = R.from_matrix(pose_matrix[:3, :3]).as_quat()
        expected = np.concatenate([expected_pos, expected_rot])
        np.testing.assert_array_almost_equal(result, expected)

    def test_pose_matrix_1d(self):
        """Test Pose with matrix format as 1D array."""
        transform = Pose(pose_type=PoseConfig.MATRIX)
        # 4x4 transformation matrix as 1D array (row-major)
        pose_matrix_1d = np.eye(4).flatten()
        result = transform(pose_matrix_1d)
        pose_matrix = pose_matrix_1d.reshape(4, 4)
        expected_pos = pose_matrix[:3, 3]
        expected_rot = R.from_matrix(pose_matrix[:3, :3]).as_quat()
        expected = np.concatenate([expected_pos, expected_rot])
        np.testing.assert_array_almost_equal(result, expected)

    def test_pose_position_orientation_quaternion(self):
        """Test Pose with position_orientation and quaternion."""
        transform = Pose(
            pose_type=PoseConfig.POSITION_ORIENTATION,
            rotation_type=RotationConfig.QUATERNION,
            angle_type=AngleConfig.RADIANS,
            seq=QuaternionOrderConfig.XYZW,
        )
        # Position + quaternion (xyzw)
        pose = np.array([1.0, 2.0, 3.0, 0.0, 0.0, 0.0, 1.0])
        result = transform(pose)
        # Should just concatenate position and quaternion (already xyzw)
        expected = np.array([1.0, 2.0, 3.0, 0.0, 0.0, 0.0, 1.0])
        np.testing.assert_array_almost_equal(result, expected)

    def test_pose_position_orientation_euler(self):
        """Test Pose with position_orientation and euler."""
        transform = Pose(
            pose_type=PoseConfig.POSITION_ORIENTATION,
            rotation_type=RotationConfig.EULER,
            angle_type=AngleConfig.RADIANS,
            seq=EulerOrderConfig.XYZ,
        )
        # Position + euler angles
        pose = np.array([1.0, 2.0, 3.0, 0.0, 0.0, 0.0])
        result = transform(pose)
        expected_pos = np.array([1.0, 2.0, 3.0])
        expected_rot = R.from_euler("xyz", [0.0, 0.0, 0.0]).as_quat()
        expected = np.concatenate([expected_pos, expected_rot])
        np.testing.assert_array_almost_equal(result, expected)

    def test_pose_unsupported_type(self):
        """Test Pose with unsupported pose type."""

        class FakePoseType:
            UNSUPPORTED = "unsupported"

        with pytest.raises(ValueError, match="Input should be"):
            Pose(pose_type=FakePoseType.UNSUPPORTED)


class TestImageFormat:
    """Tests for ImageFormat transform."""

    def test_image_format_hwc(self):
        """Test ImageFormat with HWC format (no change)."""
        transform = ImageFormat(format=ImageConventionConfig.CHANNELS_LAST)
        image = np.random.rand(32, 64, 3)  # H, W, C
        result = transform(image)
        # Test shape remains unchanged
        assert result.shape == image.shape

    def test_image_format_chw(self):
        """Test ImageFormat with CHW format (convert to HWC)."""
        transform = ImageFormat(format=ImageConventionConfig.CHANNELS_FIRST)
        image = np.random.rand(3, 32, 64)  # C, H, W
        result = transform(image)
        # Should now have shape H, W, C
        assert result.shape == (32, 64, 3)


class TestImageChannelOrder:
    """Tests for ImageChannelOrder transform."""

    def test_image_channel_order_rgb(self):
        """Test ImageChannelOrder with RGB (no change)."""
        transform = ImageChannelOrder(order=ImageChannelOrderConfig.RGB)
        image = np.random.rand(32, 64, 3)
        result = transform(image)
        np.testing.assert_array_equal(result, image)

    def test_image_channel_order_bgr(self):
        """Test ImageChannelOrder with BGR (convert to RGB)."""
        transform = ImageChannelOrder(order=ImageChannelOrderConfig.BGR)
        image = np.array([[[1.0, 2.0, 3.0]]])  # B, G, R
        result = transform(image)
        expected = np.array([[[3.0, 2.0, 1.0]]])  # R, G, B
        np.testing.assert_array_equal(result, expected)


class TestCastToNumpyDtype:
    """Tests for CastToNumpyDtype transform."""

    def test_cast_to_uint8(self):
        """Test CastToNumpyDtype to uint8."""
        transform = CastToNumpyDtype(dtype=np.uint8)
        data = np.array([1.5, 2.7, 3.9], dtype=np.float32)
        result = transform(data)
        assert result.dtype == np.uint8
        np.testing.assert_array_equal(result, np.array([1, 2, 3], dtype=np.uint8))

    def test_cast_to_float32(self):
        """Test CastToNumpyDtype to float32."""
        transform = CastToNumpyDtype(dtype=np.float32)
        data = np.array([1, 2, 3], dtype=np.int32)
        result = transform(data)
        assert result.dtype == np.float32
        np.testing.assert_array_equal(
            result, np.array([1.0, 2.0, 3.0], dtype=np.float32)
        )


class TestNumpyToScalar:
    """Tests for NumpyToScalar transform."""

    def test_numpy_to_scalar(self):
        """Test NumpyToScalar conversion."""
        transform = NumpyToScalar()
        data = np.array([42.5])
        result = transform(data)
        assert isinstance(result, (float, np.floating))
        assert result == 42.5

    def test_numpy_to_scalar_int(self):
        """Test NumpyToScalar with integer array."""
        transform = NumpyToScalar()
        data = np.array([42])
        result = transform(data)
        assert isinstance(result, (int, np.integer))
        assert result == 42


class TestScale:
    """Tests for Scale transform."""

    def test_scale_positive(self):
        """Test Scale with positive factor."""
        transform = Scale(factor=2.0)
        data = np.array([1.0, 2.0, 3.0])
        result = transform(data)
        expected = np.array([2.0, 4.0, 6.0])
        np.testing.assert_array_equal(result, expected)

    def test_scale_negative(self):
        """Test Scale with negative factor."""
        transform = Scale(factor=-1.0)
        data = np.array([1.0, 2.0, 3.0])
        result = transform(data)
        expected = np.array([-1.0, -2.0, -3.0])
        np.testing.assert_array_equal(result, expected)

    def test_scale_fractional(self):
        """Test Scale with fractional factor."""
        transform = Scale(factor=0.5)
        data = np.array([2.0, 4.0, 6.0])
        result = transform(data)
        expected = np.array([1.0, 2.0, 3.0])
        np.testing.assert_array_equal(result, expected)


class TestClip:
    """Tests for Clip transform."""

    def test_clip_within_range(self):
        """Test Clip with values within range."""
        transform = Clip(min=0.0, max=10.0)
        data = np.array([1.0, 5.0, 9.0])
        result = transform(data)
        np.testing.assert_array_equal(result, data)

    def test_clip_below_min(self):
        """Test Clip with values below minimum."""
        transform = Clip(min=0.0, max=10.0)
        data = np.array([-5.0, 1.0, 5.0])
        result = transform(data)
        expected = np.array([0.0, 1.0, 5.0])
        np.testing.assert_array_equal(result, expected)

    def test_clip_above_max(self):
        """Test Clip with values above maximum."""
        transform = Clip(min=0.0, max=10.0)
        data = np.array([1.0, 5.0, 15.0])
        result = transform(data)
        expected = np.array([1.0, 5.0, 10.0])
        np.testing.assert_array_equal(result, expected)

    def test_clip_both_bounds(self):
        """Test Clip with values outside both bounds."""
        transform = Clip(min=0.0, max=10.0)
        data = np.array([-5.0, 5.0, 15.0])
        result = transform(data)
        expected = np.array([0.0, 5.0, 10.0])
        np.testing.assert_array_equal(result, expected)


class TestNormalize:
    """Tests for Normalize transform."""

    def test_normalize_basic(self):
        """Test Normalize basic functionality."""
        transform = Normalize(min=0.0, max=10.0)
        data = np.array([0.0, 5.0, 10.0])
        result = transform(data)
        expected = np.array([0.0, 0.5, 1.0])
        np.testing.assert_array_almost_equal(result, expected)

    def test_normalize_negative_range(self):
        """Test Normalize with negative range."""
        transform = Normalize(min=-1.0, max=1.0)
        data = np.array([-1.0, 0.0, 1.0])
        result = transform(data)
        expected = np.array([0.0, 0.5, 1.0])
        np.testing.assert_array_almost_equal(result, expected)

    def test_normalize_outside_range(self):
        """Test Normalize with values outside range."""
        transform = Normalize(min=0.0, max=10.0)
        data = np.array([-5.0, 5.0, 15.0])
        result = transform(data)
        expected = np.array([-0.5, 0.5, 1.5])
        np.testing.assert_array_almost_equal(result, expected)


class TestUnnormalize:
    """Tests for Unnormalize transform."""

    def test_unnormalize_basic(self):
        """Test Unnormalize basic functionality."""
        transform = Unnormalize(min=0.0, max=10.0)
        data = np.array([0.0, 0.5, 1.0])
        result = transform(data)
        expected = np.array([0.0, 5.0, 10.0])
        np.testing.assert_array_almost_equal(result, expected)

    def test_unnormalize_negative_range(self):
        """Test Unnormalize with negative range."""
        transform = Unnormalize(min=-1.0, max=1.0)
        data = np.array([0.0, 0.5, 1.0])
        result = transform(data)
        expected = np.array([-1.0, 0.0, 1.0])
        np.testing.assert_array_almost_equal(result, expected)


class TestFlipSign:
    """Tests for FlipSign transform."""

    def test_flip_sign_positive(self):
        """Test FlipSign with positive value."""
        transform = FlipSign()
        data = 5.0
        result = transform(data)
        assert result == -5.0

    def test_flip_sign_negative(self):
        """Test FlipSign with negative value."""
        transform = FlipSign()
        data = -5.0
        result = transform(data)
        assert result == 5.0

    def test_flip_sign_zero(self):
        """Test FlipSign with zero."""
        transform = FlipSign()
        data = 0.0
        result = transform(data)
        assert result == 0.0


class TestOffset:
    """Tests for Offset transform."""

    def test_offset_positive(self):
        """Test Offset with positive value."""
        transform = Offset(value=5.0)
        data = 10.0
        result = transform(data)
        assert result == 15.0

    def test_offset_negative(self):
        """Test Offset with negative value."""
        transform = Offset(value=-5.0)
        data = 10.0
        result = transform(data)
        assert result == 5.0

    def test_offset_zero(self):
        """Test Offset with zero."""
        transform = Offset(value=0.0)
        data = 10.0
        result = transform(data)
        assert result == 10.0


class TestNanToNum:
    """Tests for NanToNum transform."""

    def test_nan_to_num_nan(self):
        """Test NanToNum with NaN values."""
        transform = NanToNum()
        data = np.array([1.0, np.nan, 3.0])
        result = transform(data)
        expected = np.array([1.0, 0.0, 3.0])
        np.testing.assert_array_equal(result, expected)

    def test_nan_to_num_inf(self):
        """Test NanToNum with infinity values."""
        transform = NanToNum()
        data = np.array([1.0, np.inf, -np.inf, 3.0])
        result = transform(data)
        expected = np.array([1.0, 0.0, 0.0, 3.0])
        np.testing.assert_array_equal(result, expected)

    def test_nan_to_num_normal(self):
        """Test NanToNum with normal values."""
        transform = NanToNum()
        data = np.array([1.0, 2.0, 3.0])
        result = transform(data)
        np.testing.assert_array_equal(result, data)


class TestDegreesToRadians:
    """Tests for DegreesToRadians transform."""

    def test_degrees_to_radians(self):
        """Test DegreesToRadians conversion."""
        transform = DegreesToRadians()
        data = np.array([0.0, 90.0, 180.0, 360.0])
        result = transform(data)
        expected = np.array([0.0, np.pi / 2, np.pi, 2 * np.pi])
        np.testing.assert_array_almost_equal(result, expected)

    def test_degrees_to_radians_negative(self):
        """Test DegreesToRadians with negative values."""
        transform = DegreesToRadians()
        data = np.array([-90.0, -180.0])
        result = transform(data)
        expected = np.array([-np.pi / 2, -np.pi])
        np.testing.assert_array_almost_equal(result, expected)


class TestLanguageFromBytes:
    """Tests for LanguageFromBytes transform."""

    def test_language_from_bytes(self):
        """Test LanguageFromBytes conversion."""
        transform = LanguageFromBytes()
        data = b"Hello, World!"
        result = transform(data)
        assert isinstance(result, str)
        assert result == "Hello, World!"

    def test_language_from_bytes_unicode(self):
        """Test LanguageFromBytes with unicode characters."""
        transform = LanguageFromBytes()
        data = "Hello, 世界!".encode()
        result = transform(data)
        assert isinstance(result, str)
        assert result == "Hello, 世界!"

    def test_language_from_bytes_empty(self):
        """Test LanguageFromBytes with empty bytes."""
        transform = LanguageFromBytes()
        data = b""
        result = transform(data)
        assert isinstance(result, str)
        assert result == ""



"""Neuracore Importer Package.

This package provides configuration classes and transformation tools for importing
and processing data to be used in Neuracore datasets.

Main components:

- config: Core enums and models for dataset and import configuration.
- data_config: Classes for mapping, formatting, and normalizing input data.
- transform: Tools for applying transformations to imported data.

Use these modules to define how your raw data is interpreted, formatted, and converted
for use in Neuracore.
"""

from neuracore_types.importer.config import *  # noqa: F403
from neuracore_types.importer.data_config import *  # noqa: F403
from neuracore_types.importer.transform import *  # noqa: F403



"""Camera data including images and camera parameters."""

from typing import Any, Literal, cast

import numpy as np
import torch
from pydantic import ConfigDict, Field, field_serializer, field_validator

from neuracore_types.batched_nc_data.batched_nc_data import BatchedNCData
from neuracore_types.nc_data.nc_data import NCData
from neuracore_types.utils.pydantic_to_ts import (
    REQUIRED_WITH_DEFAULT_FLAG,
    fix_required_with_defaults,
)

RGB_URI_PREFIX = "data:image/png;base64,"


class BatchedRGBData(BatchedNCData):
    """Batched RGB camera data."""

    type: Literal["BatchedRGBData"] = Field(
        default="BatchedRGBData", json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    frame: torch.Tensor  # (B, T, 3, H, W) uint8
    extrinsics: torch.Tensor  # (B, T, 4, 4) float16
    intrinsics: torch.Tensor  # (B, T, 3, 3) float16

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)

    @field_validator("frame", mode="before")
    @classmethod
    def decode_frame(cls, v: dict[str, Any]) -> torch.Tensor:
        """Decode frame field to torch.Tensor."""
        return cls._create_tensor_handlers("frame")[0](v)

    @field_serializer("frame", when_used="json")
    def serialize_frame(self, v: torch.Tensor) -> dict[str, Any]:
        """Serialize frame field to base64 string."""
        return self._create_tensor_handlers("frame")[1](v)

    @field_validator("extrinsics", mode="before")
    @classmethod
    def decode_extrinsics(cls, v: dict[str, Any]) -> torch.Tensor:
        """Decode extrinsics field to torch.Tensor."""
        return cls._create_tensor_handlers("extrinsics")[0](v)

    @field_serializer("extrinsics", when_used="json")
    def serialize_extrinsics(self, v: torch.Tensor) -> dict[str, Any]:
        """Serialize extrinsics field to base64 string."""
        return self._create_tensor_handlers("extrinsics")[1](v)

    @field_validator("intrinsics", mode="before")
    @classmethod
    def decode_intrinsics(cls, v: dict[str, Any]) -> torch.Tensor:
        """Decode intrinsics field to torch.Tensor."""
        return cls._create_tensor_handlers("intrinsics")[0](v)

    @field_serializer("intrinsics", when_used="json")
    def serialize_intrinsics(self, v: torch.Tensor) -> dict[str, Any]:
        """Serialize intrinsics field to base64 string."""
        return self._create_tensor_handlers("intrinsics")[1](v)

    @classmethod
    def from_nc_data(cls, nc_data: NCData) -> "BatchedNCData":
        """Create BatchedRGBData from input nc_data.

        Args:
            nc_data: NCData instance to convert

        Returns:
            BatchedNCData: Converted BatchedNCData instance
        """
        from neuracore_types.nc_data.camera_data import RGBCameraData

        rgb_data: RGBCameraData = cast(RGBCameraData, nc_data)
        # Need to change from (H, W, 3) to (3, H, W)
        frame = np.array(rgb_data.frame)
        frame = (
            torch.tensor(frame.transpose(2, 0, 1), dtype=torch.float32)
            .unsqueeze(0)
            .unsqueeze(0)
        )
        if rgb_data.extrinsics is not None:
            extrinsics = (
                torch.tensor(rgb_data.extrinsics, dtype=torch.float32)
                .unsqueeze(0)
                .unsqueeze(0)
            )
        else:
            extrinsics = torch.zeros((1, 1, 4, 4), dtype=torch.float32)
        if rgb_data.intrinsics is not None:
            intrinsics = (
                torch.tensor(rgb_data.intrinsics, dtype=torch.float32)
                .unsqueeze(0)
                .unsqueeze(0)
            )
        else:
            intrinsics = torch.zeros((1, 1, 3, 3), dtype=torch.float32)
        return cls(frame=frame, extrinsics=extrinsics, intrinsics=intrinsics)

    def transform_nc_data(self) -> None:
        """Apply in-place transformations, e.g. reshaping, reordering dimensions, etc.

        Example usage might include:
        - Reordering dimensions from (B, T, C, H, W) to (B, T, H, W, C) for images
        - Normalizing values to a specific range
        - Resizing images to a common resolution

        Args:
            nc_data: BatchedNCData instance to transform

        Returns:
            BatchedNCData: Transformed instance
        """
        # TODO: In future, the resizing spec should come from the algorithm
        #   or from user specification, not be hardcoded here.

        # Resize each frame to 224x224 while preserving (B, T, C, H, W).
        if self.frame.shape[-2:] != (224, 224):
            batch_size, time_steps, channels, _, _ = self.frame.shape
            reshaped = self.frame.reshape(
                batch_size * time_steps, channels, *self.frame.shape[-2:]
            )
            resized = torch.nn.functional.interpolate(
                reshaped, size=(224, 224), mode="bilinear", align_corners=False
            )
            self.frame = resized.reshape(batch_size, time_steps, channels, 224, 224)

    @classmethod
    def from_nc_data_list(cls, nc_data_list: list[NCData]) -> "BatchedRGBData":
        """Create BatchedRGBData from list of RGBCameraData.

        Args:
            nc_data_list: List of RGBCameraData instances to convert

        Returns:
            BatchedRGBData with shape (1, T, 3, H, W) where T = len(nc_data_list)
        """
        from neuracore_types.nc_data.camera_data import RGBCameraData

        frames = []
        extrinsics_list = []
        intrinsics_list = []

        for nc in nc_data_list:
            rgb_data: RGBCameraData = cast(RGBCameraData, nc)
            # (H, W, 3) -> (3, H, W)
            frame = np.array(rgb_data.frame).transpose(2, 0, 1)
            frames.append(frame)

            if rgb_data.extrinsics is not None:
                extrinsics_list.append(rgb_data.extrinsics)
            else:
                extrinsics_list.append(np.zeros((4, 4), dtype=np.float32))

            if rgb_data.intrinsics is not None:
                intrinsics_list.append(rgb_data.intrinsics)
            else:
                intrinsics_list.append(np.zeros((3, 3), dtype=np.float32))

        # Shape: (1, T, 3, H, W)
        frame_tensor = torch.tensor(np.stack(frames), dtype=torch.float32).unsqueeze(0)
        # Shape: (1, T, 4, 4)
        extrinsics_tensor = torch.tensor(
            np.stack(extrinsics_list), dtype=torch.float32
        ).unsqueeze(0)
        # Shape: (1, T, 3, 3)
        intrinsics_tensor = torch.tensor(
            np.stack(intrinsics_list), dtype=torch.float32
        ).unsqueeze(0)

        return cls(
            frame=frame_tensor,
            extrinsics=extrinsics_tensor,
            intrinsics=intrinsics_tensor,
        )

    @classmethod
    def sample(cls, batch_size: int = 1, time_steps: int = 1) -> "BatchedRGBData":
        """Sample an example instance of BatchedRGBData.

        Args:
            batch_size: Number of samples in the batch
            time_steps: Number of time steps in the sequence

        Returns:
            BatchedRGBData: Sampled instance
        """
        return cls(
            frame=torch.zeros(
                (batch_size, time_steps, 3, 224, 224), dtype=torch.float32
            ),
            extrinsics=torch.zeros((batch_size, time_steps, 4, 4), dtype=torch.float32),
            intrinsics=torch.zeros((batch_size, time_steps, 3, 3), dtype=torch.float32),
        )


class BatchedDepthData(BatchedNCData):
    """Batched depth camera data."""

    type: Literal["BatchedDepthData"] = Field(
        default="BatchedDepthData", json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    frame: torch.Tensor  # (B, T, 1, H, W) uint8
    extrinsics: torch.Tensor  # (B, T, 4, 4) float16
    intrinsics: torch.Tensor  # (B, T, 3, 3) float16

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)

    @field_validator("frame", mode="before")
    @classmethod
    def decode_frame(cls, v: dict[str, Any]) -> torch.Tensor:
        """Decode frame field to torch.Tensor."""
        return cls._create_tensor_handlers("frame")[0](v)

    @field_serializer("frame", when_used="json")
    def serialize_frame(self, v: torch.Tensor) -> dict[str, Any]:
        """Serialize frame field to base64 string."""
        return self._create_tensor_handlers("frame")[1](v)

    @field_validator("extrinsics", mode="before")
    @classmethod
    def decode_extrinsics(cls, v: dict[str, Any]) -> torch.Tensor:
        """Decode extrinsics field to torch.Tensor."""
        return cls._create_tensor_handlers("extrinsics")[0](v)

    @field_serializer("extrinsics", when_used="json")
    def serialize_extrinsics(self, v: torch.Tensor) -> dict[str, Any]:
        """Serialize extrinsics field to base64 string."""
        return self._create_tensor_handlers("extrinsics")[1](v)

    @field_validator("intrinsics", mode="before")
    @classmethod
    def decode_intrinsics(cls, v: dict[str, Any]) -> torch.Tensor:
        """Decode intrinsics field to torch.Tensor."""
        return cls._create_tensor_handlers("intrinsics")[0](v)

    @field_serializer("intrinsics", when_used="json")
    def serialize_intrinsics(self, v: torch.Tensor) -> dict[str, Any]:
        """Serialize intrinsics field to base64 string."""
        return self._create_tensor_handlers("intrinsics")[1](v)

    @classmethod
    def from_nc_data(cls, nc_data: NCData) -> "BatchedNCData":
        """Create BatchedDepthData from input nc_data.

        Args:
            nc_data: NCData instance to convert

        Returns:
            BatchedNCData: Converted BatchedNCData instance
        """
        from neuracore_types.nc_data.camera_data import DepthCameraData

        depth_data: DepthCameraData = cast(DepthCameraData, nc_data)
        # Need to change from (H, W) to (1, H, W)
        assert isinstance(depth_data.frame, np.ndarray)
        frame = (
            torch.tensor(depth_data.frame, dtype=torch.float32)
            .unsqueeze(0)
            .unsqueeze(0)
            .unsqueeze(0)
        )
        extrinsics = (
            torch.tensor(depth_data.extrinsics, dtype=torch.float32)
            .unsqueeze(0)
            .unsqueeze(0)
        )
        intrinsics = (
            torch.tensor(depth_data.intrinsics, dtype=torch.float32)
            .unsqueeze(0)
            .unsqueeze(0)
        )
        return cls(frame=frame, extrinsics=extrinsics, intrinsics=intrinsics)

    @classmethod
    def from_nc_data_list(cls, nc_data_list: list[NCData]) -> "BatchedDepthData":
        """Create BatchedDepthData from list of DepthCameraData.

        Args:
            nc_data_list: List of DepthCameraData instances to convert

        Returns:
            BatchedDepthData with shape (1, T, 1, H, W) where T = len(nc_data_list)
        """
        from neuracore_types.nc_data.camera_data import DepthCameraData

        frames = []
        extrinsics_list = []
        intrinsics_list = []

        for nc in nc_data_list:
            depth_data: DepthCameraData = cast(DepthCameraData, nc)
            assert isinstance(depth_data.frame, np.ndarray)
            # (H, W) -> (1, H, W)
            frames.append(depth_data.frame[np.newaxis, ...])
            extrinsics_list.append(depth_data.extrinsics)
            intrinsics_list.append(depth_data.intrinsics)

        # Shape: (1, T, 1, H, W)
        frame_tensor = torch.tensor(np.stack(frames), dtype=torch.float32).unsqueeze(0)
        # Shape: (1, T, 4, 4)
        extrinsics_tensor = torch.tensor(
            np.stack(extrinsics_list), dtype=torch.float32
        ).unsqueeze(0)
        # Shape: (1, T, 3, 3)
        intrinsics_tensor = torch.tensor(
            np.stack(intrinsics_list), dtype=torch.float32
        ).unsqueeze(0)

        return cls(
            frame=frame_tensor,
            extrinsics=extrinsics_tensor,
            intrinsics=intrinsics_tensor,
        )

    @classmethod
    def sample(cls, batch_size: int = 1, time_steps: int = 1) -> "BatchedDepthData":
        """Sample an example instance of BatchedDepthData.

        Args:
            batch_size: Number of samples in the batch
            time_steps: Number of time steps in the sequence

        Returns:
            BatchedDepthData: Sampled instance
        """
        return cls(
            frame=torch.zeros(
                (batch_size, time_steps, 1, 224, 224), dtype=torch.float32
            ),
            extrinsics=torch.zeros((batch_size, time_steps, 4, 4), dtype=torch.float32),
            intrinsics=torch.zeros((batch_size, time_steps, 3, 3), dtype=torch.float32),
        )



"""Tests for ParallelGripperOpenAmountData and BatchedParallelGripperOpenAmountData."""

from typing import cast

import numpy as np
import pytest
import torch

from neuracore_types import (
    BatchedParallelGripperOpenAmountData,
    ParallelGripperOpenAmountData,
)
from neuracore_types.batched_nc_data import DATA_TYPE_TO_BATCHED_NC_DATA_CLASS
from neuracore_types.importer.config import NormalizeConfig
from neuracore_types.importer.data_config import DataFormat, MappingItem
from neuracore_types.importer.transform import Clip, Normalize
from neuracore_types.nc_data import DATA_TYPE_TO_NC_DATA_CLASS, DataType
from neuracore_types.nc_data.parallel_gripper_open_amount_data import (
    ParallelGripperOpenAmountDataImportConfig,
)


class TestParallelGripperOpenAmountData:
    """Tests for ParallelGripperOpenAmountData functionality."""

    def test_sample(self):
        """Test ParallelGripperOpenAmountData.sample() creates valid instance."""
        data = ParallelGripperOpenAmountData.sample()
        assert isinstance(data, ParallelGripperOpenAmountData)
        assert isinstance(data.open_amount, float)
        assert data.type == "ParallelGripperOpenAmountData"

    def test_calculate_statistics(self):
        """Test calculate_statistics() returns valid stats."""
        data = ParallelGripperOpenAmountData(open_amount=0.5)
        stats = data.calculate_statistics()

        assert stats.type == "ParallelGripperOpenAmountDataStats"
        assert stats.open_amount is not None
        assert len(stats.open_amount.mean) == 1
        assert stats.open_amount.mean[0] == 0.5

    def test_value_range_validation(self):
        """Test that extreme values are handled correctly."""
        data = ParallelGripperOpenAmountData(open_amount=1.0)
        assert data.open_amount == 1.0

        data = ParallelGripperOpenAmountData(open_amount=0.0)
        assert data.open_amount == 0.0

    def test_negative_gripper_value(self):
        """Test handling of negative gripper values."""
        data = ParallelGripperOpenAmountData(open_amount=-0.5)
        assert data.open_amount == -0.5

    def test_serialization(self):
        """Test JSON serialization and deserialization."""
        data = ParallelGripperOpenAmountData(open_amount=0.75)
        json_str = data.model_dump_json()
        loaded = ParallelGripperOpenAmountData.model_validate_json(json_str)

        assert loaded.open_amount == data.open_amount
        assert loaded.timestamp == data.timestamp

    def test_invalid_type(self):
        """Test that invalid value type raises error."""
        with pytest.raises(Exception):
            ParallelGripperOpenAmountData(open_amount="not a number")


class TestBatchedParallelGripperOpenAmountData:
    """Tests for BatchedParallelGripperOpenAmountData functionality."""

    def test_from_nc_data(self):
        """Test BatchedParallelGripperOpenAmountData.from_nc_data()."""
        gripper_data = ParallelGripperOpenAmountData(open_amount=0.75)
        batched = BatchedParallelGripperOpenAmountData.from_nc_data(gripper_data)

        assert isinstance(batched, BatchedParallelGripperOpenAmountData)
        assert batched.open_amount.shape == (1, 1, 1)
        assert batched.open_amount[0, 0, 0] == 0.75

    def test_sample(self):
        """Test BatchedParallelGripperOpenAmountData.sample()."""
        batched = BatchedParallelGripperOpenAmountData.sample(
            batch_size=3, time_steps=2
        )
        assert batched.open_amount.shape == (3, 2, 1)

    def test_sample_large_dimensions(self):
        """Test sample with large dimensions."""
        batched = BatchedParallelGripperOpenAmountData.sample(
            batch_size=10, time_steps=20
        )
        assert batched.open_amount.shape == (10, 20, 1)

    def test_to_device(self):
        """Test moving to different device."""
        batched = BatchedParallelGripperOpenAmountData.sample(
            batch_size=2, time_steps=3
        )
        batched_cpu = batched.to(torch.device("cpu"))
        batched_cpu = cast(BatchedParallelGripperOpenAmountData, batched_cpu)

        assert batched_cpu.open_amount.device.type == "cpu"
        assert torch.allclose(batched_cpu.open_amount, batched.open_amount)

    def test_from_nc_data_preserves_value(self):
        """Test that from_nc_data preserves exact value."""
        test_value = 0.12345
        gripper_data = ParallelGripperOpenAmountData(open_amount=test_value)
        batched = BatchedParallelGripperOpenAmountData.from_nc_data(gripper_data)
        batched = cast(BatchedParallelGripperOpenAmountData, batched)

        assert torch.isclose(
            batched.open_amount[0, 0, 0], torch.tensor(test_value), rtol=1e-5
        )

    def test_can_serialize_deserialize(self):
        """Test JSON serialization and deserialization."""
        batched = BatchedParallelGripperOpenAmountData.sample(
            batch_size=2, time_steps=2
        )
        json_str = batched.model_dump_json()
        loaded = BatchedParallelGripperOpenAmountData.model_validate_json(json_str)

        assert torch.equal(loaded.open_amount, batched.open_amount)
        assert loaded.open_amount.shape == batched.open_amount.shape


class TestParallelGripperStatistics:
    """Tests for ParallelGripperOpenAmountData statistics."""

    def test_statistics_values(self):
        """Test that statistics contain correct values."""
        data = ParallelGripperOpenAmountData(open_amount=0.8)
        stats = data.calculate_statistics()

        assert np.isclose(stats.open_amount.mean[0], 0.8)
        assert stats.open_amount.count[0] == 1
        assert np.isclose(stats.open_amount.min[0], 0.8)
        assert np.isclose(stats.open_amount.max[0], 0.8)

    def test_statistics_concatenation(self):
        """Test that gripper statistics can be concatenated."""
        data1 = ParallelGripperOpenAmountData(open_amount=0.3)
        data2 = ParallelGripperOpenAmountData(open_amount=0.7)

        stats1 = data1.calculate_statistics()
        stats2 = data2.calculate_statistics()

        concatenated = stats1.open_amount.concatenate(stats2.open_amount)
        assert len(concatenated.mean) == 2
        assert np.isclose(concatenated.mean[0], 0.3)
        assert np.isclose(concatenated.mean[1], 0.7)


class TestParallelGripperTargetOpenAmountDataType:
    """Tests for PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS DataType mapping."""

    def test_target_data_type_exists(self):
        """Test that PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS DataType exists."""
        assert hasattr(DataType, "PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS")
        assert (
            DataType.PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS.value
            == "PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS"
        )

    def test_target_data_type_mapped_to_nc_data_class(self):
        """Test that target DataType is mapped to ParallelGripperOpenAmountData."""
        assert (
            DataType.PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS in DATA_TYPE_TO_NC_DATA_CLASS
        )
        assert (
            DATA_TYPE_TO_NC_DATA_CLASS[DataType.PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS]
            == ParallelGripperOpenAmountData
        )

    def test_target_data_type_mapped_to_batched_nc_data_class(self):
        """Test target DataType is mapped to BatchedParallelGripperOpenAmountData."""
        assert (
            DataType.PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS
            in DATA_TYPE_TO_BATCHED_NC_DATA_CLASS
        )
        assert (
            DATA_TYPE_TO_BATCHED_NC_DATA_CLASS[
                DataType.PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS
            ]
            == BatchedParallelGripperOpenAmountData
        )

    def test_target_uses_same_data_class_as_state(self):
        """Test that target and state use the same underlying data classes."""
        # Similar to how JOINT_TARGET_POSITIONS uses JointData
        assert (
            DATA_TYPE_TO_NC_DATA_CLASS[DataType.PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS]
            == DATA_TYPE_TO_NC_DATA_CLASS[DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS]
        )
        assert (
            DATA_TYPE_TO_BATCHED_NC_DATA_CLASS[
                DataType.PARALLEL_GRIPPER_TARGET_OPEN_AMOUNTS
            ]
            == DATA_TYPE_TO_BATCHED_NC_DATA_CLASS[
                DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS
            ]
        )


class TestParallelGripperOpenAmountDataImportConfig:
    """Tests for ParallelGripperOpenAmountDataImportConfig class."""

    def test_parallel_gripper_transforms_no_normalize(self):
        """Test ParallelGripperOpenAmountDataImportConfig transforms w/o normalize."""
        data_point = ParallelGripperOpenAmountDataImportConfig(
            source="gripper",
            mapping=[MappingItem(name="gripper_open")],
            format=DataFormat(normalize=None),
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert not any(isinstance(t, Normalize) for t in transforms)
        assert any(isinstance(t, Clip) for t in transforms)

    def test_parallel_gripper_transforms_with_normalize(self):
        """Test ParallelGripperOpenAmountDataImportConfig transforms with normalize."""
        normalize = NormalizeConfig(min=0.0, max=100.0)
        data_point = ParallelGripperOpenAmountDataImportConfig(
            source="gripper",
            mapping=[MappingItem(name="gripper_open")],
            format=DataFormat(normalize=normalize),
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert any(isinstance(t, Normalize) for t in transforms)



"""Base classes for Neuracore data types."""

from typing import Any, Callable, Union

import torch
from pydantic import BaseModel, ConfigDict

from neuracore_types.nc_data.nc_data import NCData


class BatchedNCData(BaseModel):
    """Base class for batched Neuracore data."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def to(self, device: torch.device) -> "BatchedNCData":
        """Move all tensors to the specified device."""
        data_dict = self.model_dump()
        moved_data = {
            key: value.to(device) if isinstance(value, torch.Tensor) else value
            for key, value in data_dict.items()
        }
        return self.__class__(**moved_data)

    @classmethod
    def from_nc_data(cls, nc_data: NCData) -> "BatchedNCData":
        """Create BatchedNCData from NCData by adding time and batch dimensions."""
        raise NotImplementedError(
            "from_nc_data method must be implemented in subclasses."
        )

    @classmethod
    def from_nc_data_list(cls, nc_data_list: list[NCData]) -> "BatchedNCData":
        """Create BatchedNCData from list of NCData, stacking along time dimension.

        This method is more efficient than calling from_nc_data() multiple times
        and concatenating, as it creates a single tensor directly.

        Args:
            nc_data_list: List of NCData instances to convert

        Returns:
            BatchedNCData with shape (1, T, ...) where T = len(nc_data_list)
        """
        raise NotImplementedError(
            "from_nc_data_list method must be implemented in subclasses."
        )

    def transform_nc_data(self) -> None:
        """Apply in-place transformations, e.g. reshaping, reordering dimensions, etc.

        Example usage might include:
        - Reordering dimensions from (B, T, C, H, W) to (B, T, H, W, C) for images
        - Normalizing values to a specific range
        - Resizing images to a common resolution

        Args:
            nc_data: BatchedNCData instance to transform

        Returns:
            BatchedNCData: Transformed instance
        """
        pass

    @classmethod
    def sample(cls, batch_size: int = 1, time_steps: int = 1) -> "BatchedNCData":
        """Sample an example instance of BatchedNCData."""
        raise NotImplementedError("sample method must be implemented in subclasses.")

    @staticmethod
    def _create_tensor_handlers(
        field_name: str,
    ) -> tuple[
        Callable[[Union[dict[str, Any], torch.Tensor]], torch.Tensor],
        Callable[[torch.Tensor], dict[str, Any]],
    ]:
        """Create validator and serializer for a torch.Tensor field."""

        def validator(v: Union[dict[str, Any], torch.Tensor]) -> torch.Tensor:
            if isinstance(v, torch.Tensor):
                return v
            elif isinstance(v, dict) and v.get("_tensor_encoded"):
                # Convert list back to tensor with proper dtype and shape
                dtype_str = v["dtype"]
                # Map numpy dtype strings to torch dtypes
                dtype_map = {
                    "float32": torch.float32,
                    "float64": torch.float64,
                    "float16": torch.float16,
                    "int32": torch.int32,
                    "int64": torch.int64,
                    "int16": torch.int16,
                    "int8": torch.int8,
                    "uint8": torch.uint8,
                    "bool": torch.bool,
                }
                torch_dtype = dtype_map.get(dtype_str, torch.float32)
                return torch.tensor(v["data"], dtype=torch_dtype).reshape(v["shape"])
            else:
                raise ValueError(f"Invalid value for {field_name}")

        def serializer(v: torch.Tensor) -> dict[str, Any]:
            return {
                "_tensor_encoded": True,
                "shape": list(v.shape),
                "dtype": str(v.dtype).replace("torch.", ""),  # e.g., "float32"
                "data": v.cpu().numpy().tolist(),
            }

        return validator, serializer



"""Configuration classes for specifying the format of the input data.

This module defines configuration classes to specify the format of the input
data when importing it into Neuracore.

"""

from pydantic import BaseModel, ConfigDict, Field, model_validator

from neuracore_types.importer.config import (
    AngleConfig,
    DistanceUnitsConfig,
    ImageChannelOrderConfig,
    ImageConventionConfig,
    IndexRangeConfig,
    JointPositionTypeConfig,
    LanguageConfig,
    NormalizeConfig,
    OrientationConfig,
    PoseConfig,
    TorqueUnitsConfig,
    VisualJointTypeConfig,
)
from neuracore_types.importer.transform import DataTransformSequence


class MappingItem(BaseModel):
    """Mapping information to extract individual items from a data source."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str
    source_name: str | None = None
    index: int | None = None
    index_range: IndexRangeConfig | None = None
    offset: float = 0.0
    inverted: bool = False
    transforms: DataTransformSequence = Field(
        default_factory=lambda: DataTransformSequence(transforms=[])
    )

    @model_validator(mode="after")
    def validate_index_provided(self) -> "MappingItem":
        """Validate that index an index_range are not provided together."""
        if self.index is not None and self.index_range is not None:
            raise ValueError("index and index_range cannot be provided together")
        return self


class DataFormat(BaseModel):
    """Per datatype format specifications.

    This class is used for the 'format:' section in YAML config files and
    when importing data. Relevant fields depend on the data type.
    """

    # RGB image format fields
    image_convention: ImageConventionConfig = ImageConventionConfig.CHANNELS_LAST
    order_of_channels: ImageChannelOrderConfig = ImageChannelOrderConfig.RGB
    normalized_pixel_values: bool = False

    # Units format fields
    angle_units: AngleConfig = AngleConfig.RADIANS
    torque_units: TorqueUnitsConfig = TorqueUnitsConfig.NM
    distance_units: DistanceUnitsConfig = DistanceUnitsConfig.M

    # Pose format fields
    pose_type: PoseConfig = PoseConfig.MATRIX
    orientation: OrientationConfig | None = None

    # Language format fields
    language_type: LanguageConfig = LanguageConfig.STRING

    # Normalize format fields
    normalize: NormalizeConfig | None = None

    # Gripper amount format fields
    invert_gripper_amount: bool = False

    # Visual joint type fields
    visual_joint_type: VisualJointTypeConfig = VisualJointTypeConfig.CUSTOM

    # Joint position type fields
    joint_position_type: JointPositionTypeConfig = JointPositionTypeConfig.CUSTOM
    ik_init_config: list[float] | None = None



"""Base classes for Neuracore data types."""

import time
from typing import Any, Optional, Union

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator

from neuracore_types.importer.data_config import DataFormat, MappingItem
from neuracore_types.utils.pydantic_to_ts import (
    REQUIRED_WITH_DEFAULT_FLAG,
    fix_required_with_defaults,
)


class NCDataStats(BaseModel):
    """Base class for statistics of Neuracore data types."""

    pass


class NCDataImportConfig(BaseModel):
    """Configuration for importing data to Neuracore."""

    source: str = Field(
        default="",
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )
    mapping: list[MappingItem] = Field(
        default_factory=list,
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )
    format: DataFormat = Field(
        default_factory=DataFormat,
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )

    def __init__(self, **data: Any) -> None:
        """Initialize the NCDataImportConfig."""
        super().__init__(**data)
        self._populate_transforms()

    def _populate_transforms(self) -> None:
        """Populate transforms based on configuration."""
        raise NotImplementedError(
            "Subclasses must implement _populate_transforms() method."
        )


class NCData(BaseModel):
    """Base class for all Neuracore data with automatic timestamping.

    Provides a common base for all data types in the system with automatic
    timestamp generation for temporal synchronization and data ordering.
    """

    timestamp: float = Field(
        default_factory=lambda: time.time(),
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)

    def calculate_statistics(self) -> NCDataStats:
        """Calculate the statistics for this data type."""
        raise NotImplementedError(
            "Subclasses must implement calculate_statistics() method."
        )

    @classmethod
    def sample(cls) -> "NCData":
        """Sample an example NCData instance."""
        raise NotImplementedError("sample method must be implemented in subclasses.")


class DataItemStats(BaseModel):
    """Statistical summary of data dimensions and distributions.

    Contains statistical information about data arrays including means,
    standard deviations, counts, and maximum lengths for normalization
    and model configuration purposes.
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True, json_schema_extra=fix_required_with_defaults
    )

    mean: np.ndarray = Field(
        default_factory=lambda: np.array([]),
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )
    std: np.ndarray = Field(
        default_factory=lambda: np.array([]),
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )
    count: np.ndarray = Field(
        default_factory=lambda: np.array([]),
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )
    min: np.ndarray = Field(
        default_factory=lambda: np.array([]),
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )
    max: np.ndarray = Field(
        default_factory=lambda: np.array([]),
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )

    def concatenate(self, other: "DataItemStats") -> "DataItemStats":
        """Concatenate two DataItemStats objects along the data dimension."""
        if not isinstance(other, DataItemStats):
            raise ValueError("Can only concatenate with another DataItemStats object.")

        return DataItemStats(
            mean=np.concatenate((self.mean, other.mean)),
            std=np.concatenate((self.std, other.std)),
            count=np.concatenate((self.count, other.count)),
            min=np.concatenate((self.min, other.min)),
            max=np.concatenate((self.max, other.max)),
        )

    @classmethod
    def _decode_field(
        cls, v: Union[list, np.ndarray], dtype: Any
    ) -> Optional[np.ndarray]:
        """Decode field to NumPy array with specified dtype."""
        return np.array(v, dtype=dtype) if isinstance(v, list) else v

    @staticmethod
    def _serialize_field(v: Optional[np.ndarray]) -> Optional[list]:
        """Serialize NumPy array to list."""
        return v.tolist() if v is not None else None

    @field_validator("mean", mode="before")
    @classmethod
    def decode_mean(cls, v: Union[list, np.ndarray]) -> Optional[np.ndarray]:
        """Decode mean field to NumPy array."""
        return cls._decode_field(v, np.float32)

    @field_serializer("mean", when_used="json")
    def serialize_mean(self, v: Optional[np.ndarray]) -> Optional[list]:
        """Serialize mean field to JSON list."""
        return self._serialize_field(v)

    @field_validator("std", mode="before")
    @classmethod
    def decode_std(cls, v: Union[list, np.ndarray]) -> Optional[np.ndarray]:
        """Decode std field to NumPy array."""
        return cls._decode_field(v, np.float32)

    @field_serializer("std", when_used="json")
    def serialize_std(self, v: Optional[np.ndarray]) -> Optional[list]:
        """Serialize std field to JSON list."""
        return self._serialize_field(v)

    @field_validator("count", mode="before")
    @classmethod
    def decode_count(cls, v: Union[list, np.ndarray]) -> Optional[np.ndarray]:
        """Decode count field to NumPy array."""
        return cls._decode_field(v, np.int64)

    @field_serializer("count", when_used="json")
    def serialize_count(self, v: Optional[np.ndarray]) -> Optional[list]:
        """Serialize count field to JSON list."""
        return self._serialize_field(v)

    @field_validator("min", mode="before")
    @classmethod
    def decode_min(cls, v: Union[list, np.ndarray]) -> Optional[np.ndarray]:
        """Decode min field to NumPy array."""
        return cls._decode_field(v, np.float32)

    @field_serializer("min", when_used="json")
    def serialize_min(self, v: Optional[np.ndarray]) -> Optional[list]:
        """Serialize min field to JSON list."""
        return self._serialize_field(v)

    @field_validator("max", mode="before")
    @classmethod
    def decode_max(cls, v: Union[list, np.ndarray]) -> Optional[np.ndarray]:
        """Decode max field to NumPy array."""
        return cls._decode_field(v, np.float32)

    @field_serializer("max", when_used="json")
    def serialize_max(self, v: Optional[np.ndarray]) -> Optional[list]:
        """Serialize max field to JSON list."""
        return self._serialize_field(v)



"""Pose data types for 6DOF poses."""

from typing import Any, Literal, cast

import torch
from pydantic import ConfigDict, Field, field_serializer, field_validator

from neuracore_types.batched_nc_data.batched_nc_data import BatchedNCData
from neuracore_types.nc_data.nc_data import NCData
from neuracore_types.utils.pydantic_to_ts import (
    REQUIRED_WITH_DEFAULT_FLAG,
    fix_required_with_defaults,
)


class BatchedPoseData(BatchedNCData):
    """Batched pose data for multiple time steps and batch sizes."""

    type: Literal["BatchedPoseData"] = Field(
        default="BatchedPoseData", json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    pose: torch.Tensor  # (B, T, 7) float32

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)

    @field_validator("pose", mode="before")
    @classmethod
    def decode_pose(cls, v: dict[str, Any]) -> torch.Tensor:
        """Decode pose field to torch.Tensor."""
        return cls._create_tensor_handlers("pose")[0](v)

    @field_serializer("pose", when_used="json")
    def serialize_pose(self, v: torch.Tensor) -> dict[str, Any]:
        """Serialize pose field to base64 string."""
        return self._create_tensor_handlers("pose")[1](v)

    @classmethod
    def from_nc_data(cls, nc_data: NCData) -> "BatchedNCData":
        """Create BatchedPoseData from PoseData.

        Args:
            nc_data: Input NCData instance

        Returns:
            BatchedNCData: BatchedPoseData instance
        """
        from neuracore_types.nc_data.pose_data import PoseData

        pose_data: PoseData = cast(PoseData, nc_data)
        pose = (
            torch.tensor(pose_data.pose, dtype=torch.float32).unsqueeze(0).unsqueeze(0)
        )
        return cls(pose=pose)

    @classmethod
    def from_nc_data_list(cls, nc_data_list: list[NCData]) -> "BatchedPoseData":
        """Create BatchedPoseData from list of PoseData.

        Args:
            nc_data_list: List of PoseData instances to convert

        Returns:
            BatchedPoseData with shape (1, T, 7) where T = len(nc_data_list)
        """
        from neuracore_types.nc_data.pose_data import PoseData

        poses = [cast(PoseData, nc).pose for nc in nc_data_list]
        # Shape: (1, T, 7)
        pose_tensor = torch.tensor(poses, dtype=torch.float32).unsqueeze(0)
        return cls(pose=pose_tensor)

    @classmethod
    def sample(cls, batch_size: int = 1, time_steps: int = 1) -> "BatchedPoseData":
        """Sample an example instance of BatchedPoseData.

        Args:
            batch_size: Number of samples in the batch
            time_steps: Number of time steps in the sequence

        Returns:
            BatchedPoseData: Sampled BatchedPoseData instance
        """
        return cls(pose=torch.zeros((batch_size, time_steps, 7), dtype=torch.float32))



"""Endpoint request models for deploying models."""

from pydantic import BaseModel, Field

from neuracore_types.episode import DataSpec
from neuracore_types.training.training import GPUType


class DeploymentConfig(BaseModel):
    """Configuration for model deployment.

    Attributes:
        machine_type: Type of machine to use for deployment.
            If None, the backend will auto-determine based on gpu_type.
        gpu_type: Type of GPU to use for deployment.
        gpu_count: Number of GPUs to use for deployment.
    """

    machine_type: str | None = None
    gpu_type: GPUType | None = GPUType.NVIDIA_TESLA_T4
    gpu_count: int | None = 1


class DeploymentRequest(BaseModel):
    """Request model for deploying a model.

    Attributes:
        training_id: Identifier of the trained model to deploy.
        name: Optional name for the endpoint.
        ttl: Optional time-to-live in seconds for the endpoint.
        model_input_order: Specification of the model input data order.
        model_output_order: Specification of the model output data order.
        config: Deployment configuration parameters.
    """

    training_id: str
    name: str | None = None
    ttl: int | None = None
    model_input_order: DataSpec
    model_output_order: DataSpec
    config: DeploymentConfig = Field(default_factory=DeploymentConfig)



"""Tests for JointData and BatchedJointData."""

import pytest
import torch

from neuracore_types import BatchedJointData, JointData
from neuracore_types.batched_nc_data import DATA_TYPE_TO_BATCHED_NC_DATA_CLASS
from neuracore_types.importer.config import (
    AngleConfig,
    TorqueUnitsConfig,
    VisualJointTypeConfig,
)
from neuracore_types.importer.data_config import DataFormat, MappingItem
from neuracore_types.importer.transform import (
    Clip,
    DegreesToRadians,
    FlipSign,
    NumpyToScalar,
    Offset,
    Scale,
    Unnormalize,
)
from neuracore_types.nc_data import DATA_TYPE_TO_NC_DATA_CLASS, DataType
from neuracore_types.nc_data.joint_data import (
    JointPositionsDataImportConfig,
    JointTorquesDataImportConfig,
    JointVelocitiesDataImportConfig,
    VisualJointPositionsDataImportConfig,
)


class TestJointData:
    """Tests for JointData functionality."""

    def test_sample(self):
        """Test JointData.sample() creates valid instance."""
        data = JointData.sample()
        assert isinstance(data, JointData)
        assert isinstance(data.value, float)
        assert data.value == 0.0
        assert data.type == "JointData"

    def test_calculate_statistics(self):
        """Test calculate_statistics() for joint data."""
        data = JointData(value=1.5)
        stats = data.calculate_statistics()

        assert stats.type == "JointDataStats"
        assert stats.value is not None
        assert stats.value.mean[0] == 1.5
        assert stats.value.count[0] == 1

    def test_serialization(self):
        """Test JSON serialization and deserialization."""
        data = JointData(value=2.5)
        json_str = data.model_dump_json()
        loaded = JointData.model_validate_json(json_str)

        assert loaded.value == data.value
        assert loaded.timestamp == data.timestamp

    def test_invalid_type(self):
        """Test that invalid joint value type raises error."""
        with pytest.raises(Exception):
            JointData(value="not a number")

    def test_negative_value(self):
        """Test handling of negative joint values."""
        data = JointData(value=-1.5)
        assert data.value == -1.5

    def test_zero_value(self):
        """Test handling of zero joint value."""
        data = JointData(value=0.0)
        assert data.value == 0.0

    def test_large_value(self):
        """Test handling of large joint values."""
        data = JointData(value=1000.0)
        assert data.value == 1000.0


class TestBatchedJointData:
    """Tests for BatchedJointData functionality."""

    def test_from_nc_data(self):
        """Test BatchedJointData.from_nc_data() conversion."""
        joint_data = JointData(value=2.5)
        batched = BatchedJointData.from_nc_data(joint_data)

        assert isinstance(batched, BatchedJointData)
        assert batched.value.shape == (1, 1, 1)
        assert batched.value[0, 0, 0] == 2.5

    def test_sample(self):
        """Test BatchedJointData.sample() with different dimensions."""
        batched = BatchedJointData.sample(batch_size=4, time_steps=6)
        assert batched.value.shape == (4, 6, 1)

    def test_sample_single_dimension(self):
        """Test sample with single batch and timestep."""
        batched = BatchedJointData.sample(batch_size=1, time_steps=1)
        assert batched.value.shape == (1, 1, 1)

    def test_sample_large_dimensions(self):
        """Test sample with large dimensions."""
        batched = BatchedJointData.sample(batch_size=100, time_steps=50)
        assert batched.value.shape == (100, 50, 1)

    def test_to_device(self):
        """Test moving BatchedJointData to different device."""
        batched = BatchedJointData.sample(batch_size=2, time_steps=3)
        batched_cpu = batched.to(torch.device("cpu"))

        assert batched_cpu.value.device.type == "cpu"
        assert batched_cpu.value.shape == batched.value.shape

    def test_from_nc_data_preserves_value(self):
        """Test that from_nc_data preserves exact value."""
        test_value = 3.14159
        joint_data = JointData(value=test_value)
        batched = BatchedJointData.from_nc_data(joint_data)

        assert torch.isclose(
            batched.value[0, 0, 0], torch.tensor(test_value), rtol=1e-5
        )

    def test_can_serialize_deserialize(self):
        """Test JSON serialization and deserialization of BatchedJointData."""
        batched = BatchedJointData.sample(batch_size=2, time_steps=2)
        json_str = batched.model_dump_json()
        loaded = BatchedJointData.model_validate_json(json_str)

        assert torch.equal(loaded.value, batched.value)
        assert loaded.value.shape == batched.value.shape

    def test_from_nc_data_list_single_item(self):
        """Test from_nc_data_list with single joint value."""
        joint_data = JointData(value=1.5)
        batched = BatchedJointData.from_nc_data_list([joint_data])

        assert isinstance(batched, BatchedJointData)
        assert batched.value.shape == (1, 1, 1)
        assert batched.value[0, 0, 0] == 1.5

    def test_from_nc_data_list_multiple_items(self):
        """Test from_nc_data_list with multiple joint values."""
        values = [0.5, 1.0, 1.5, 2.0, 2.5]
        joint_data_list = [JointData(value=v) for v in values]
        batched = BatchedJointData.from_nc_data_list(joint_data_list)

        assert batched.value.shape == (1, 5, 1)
        for i, expected_val in enumerate(values):
            assert torch.isclose(
                batched.value[0, i, 0], torch.tensor(expected_val), rtol=1e-5
            )

    def test_from_nc_data_list_large_batch(self):
        """Test from_nc_data_list with large number of joint values."""
        num_joints = 100
        joint_data_list = [JointData(value=float(i)) for i in range(num_joints)]
        batched = BatchedJointData.from_nc_data_list(joint_data_list)

        assert batched.value.shape == (1, num_joints, 1)
        assert batched.value[0, 0, 0] == 0.0
        assert batched.value[0, 99, 0] == 99.0

    def test_from_nc_data_list_preserves_order(self):
        """Test that from_nc_data_list preserves order of joint values."""
        values = [3.14, -1.5, 0.0, 2.71, -5.0]
        joint_data_list = [JointData(value=v) for v in values]
        batched = BatchedJointData.from_nc_data_list(joint_data_list)

        for i, expected_val in enumerate(values):
            assert torch.isclose(
                batched.value[0, i, 0], torch.tensor(expected_val), rtol=1e-5
            )

    def test_from_nc_data_list_negative_values(self):
        """Test from_nc_data_list with negative joint values."""
        joint_data_list = [JointData(value=-1.0), JointData(value=-2.0)]
        batched = BatchedJointData.from_nc_data_list(joint_data_list)

        assert batched.value[0, 0, 0] == -1.0
        assert batched.value[0, 1, 0] == -2.0


class TestJointDataStatistics:
    """Tests for JointData statistics."""

    def test_statistics_values(self):
        """Test that statistics contain correct values."""
        data = JointData(value=5.0)
        stats = data.calculate_statistics()

        assert stats.value.mean[0] == 5.0
        assert stats.value.count[0] == 1
        assert stats.value.min[0] == 5.0
        assert stats.value.max[0] == 5.0
        assert stats.value.std[0] == 0.0

    def test_statistics_concatenation(self):
        """Test that joint statistics can be concatenated."""
        data1 = JointData(value=1.0)
        data2 = JointData(value=2.0)

        stats1 = data1.calculate_statistics()
        stats2 = data2.calculate_statistics()

        concatenated = stats1.value.concatenate(stats2.value)
        assert len(concatenated.mean) == 2
        assert concatenated.mean[0] == 1.0
        assert concatenated.mean[1] == 2.0

    def test_statistics_with_negative_value(self):
        """Test statistics with negative joint value."""
        data = JointData(value=-3.5)
        stats = data.calculate_statistics()

        assert stats.value.mean[0] == -3.5
        assert stats.value.min[0] == -3.5
        assert stats.value.max[0] == -3.5


class TestJointPositionsDataImportConfig:
    """Tests for JointPositionsDataImportConfig class."""

    def test_joint_positions_auto_index(self):
        """Test JointPositionsDataImportConfig auto-assigns indexes."""
        data_point = JointPositionsDataImportConfig(
            source="joints",
            mapping=[
                MappingItem(name="joint_0"),
                MappingItem(name="joint_1"),
                MappingItem(name="joint_2"),
            ],
        )
        assert data_point.mapping[0].index == 0
        assert data_point.mapping[1].index == 1
        assert data_point.mapping[2].index == 2

    def test_joint_positions_all_indexes_provided(self):
        """Test JointPositionsDataImportConfig with all indexes provided."""
        data_point = JointPositionsDataImportConfig(
            source="joints",
            mapping=[
                MappingItem(name="joint_0", index=5),
                MappingItem(name="joint_1", index=6),
            ],
        )
        assert data_point.mapping[0].index == 5
        assert data_point.mapping[1].index == 6

    def test_joint_positions_mixed_indexes_invalid(self):
        """Test JointPositionsDataImportConfig validation with mixed indexes."""
        with pytest.raises(
            ValueError,
            match="All or none of the mapping items in",
        ):
            JointPositionsDataImportConfig(
                source="joints",
                mapping=[
                    MappingItem(name="joint_0", index=0),
                    MappingItem(name="joint_1"),  # Missing index
                ],
            )

    def test_joint_positions_transforms_radians(self):
        """Test JointPositionsDataImportConfig transforms for radians."""
        data_point = JointPositionsDataImportConfig(
            source="joints",
            mapping=[MappingItem(name="joint_0")],
            format=DataFormat(angle_units=AngleConfig.RADIANS),
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert not any(isinstance(t, DegreesToRadians) for t in transforms)
        assert isinstance(transforms[-1], NumpyToScalar)

    def test_joint_positions_transforms_degrees(self):
        """Test JointPositionsDataImportConfig transforms for degrees."""
        data_point = JointPositionsDataImportConfig(
            source="joints",
            mapping=[MappingItem(name="joint_0")],
            format=DataFormat(angle_units=AngleConfig.DEGREES),
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert any(isinstance(t, DegreesToRadians) for t in transforms)

    def test_joint_positions_transforms_inverted(self):
        """Test JointPositionsDataImportConfig transforms with inverted flag."""
        data_point = JointPositionsDataImportConfig(
            source="joints",
            mapping=[MappingItem(name="joint_0", inverted=True)],
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert any(isinstance(t, FlipSign) for t in transforms)

    def test_joint_positions_transforms_offset(self):
        """Test JointPositionsDataImportConfig transforms with offset."""
        data_point = JointPositionsDataImportConfig(
            source="joints",
            mapping=[MappingItem(name="joint_0", offset=1.5)],
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert any(isinstance(t, Offset) for t in transforms)


class TestJointVelocitiesDataImportConfig:
    """Tests for JointVelocitiesDataImportConfig class."""

    def test_joint_velocities_auto_index(self):
        """Test JointVelocitiesDataImportConfig auto-assigns indexes."""
        data_point = JointVelocitiesDataImportConfig(
            source="joints",
            mapping=[MappingItem(name="joint_0"), MappingItem(name="joint_1")],
        )
        assert data_point.mapping[0].index == 0
        assert data_point.mapping[1].index == 1

    def test_joint_velocities_transforms_degrees(self):
        """Test JointVelocitiesDataImportConfig transforms for degrees."""
        data_point = JointVelocitiesDataImportConfig(
            source="joints",
            mapping=[MappingItem(name="joint_0")],
            format=DataFormat(angle_units=AngleConfig.DEGREES),
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert any(isinstance(t, DegreesToRadians) for t in transforms)


class TestJointTorquesDataImportConfig:
    """Tests for JointTorquesDataImportConfig class."""

    def test_joint_torques_transforms_nm(self):
        """Test JointTorquesDataImportConfig transforms for Nm."""
        data_point = JointTorquesDataImportConfig(
            source="joints",
            mapping=[MappingItem(name="joint_0")],
            format=DataFormat(torque_units=TorqueUnitsConfig.NM),
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert not any(isinstance(t, Scale) for t in transforms)

    def test_joint_torques_transforms_ncm(self):
        """Test JointTorquesDataImportConfig transforms for Ncm."""
        data_point = JointTorquesDataImportConfig(
            source="joints",
            mapping=[MappingItem(name="joint_0")],
            format=DataFormat(torque_units=TorqueUnitsConfig.NCM),
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert any(isinstance(t, Scale) for t in transforms)


class TestVisualJointPositionsDataType:
    """Tests for VISUAL_JOINT_POSITIONS DataType mapping."""

    def test_visual_joint_positions_data_type_exists(self):
        """Test that VISUAL_JOINT_POSITIONS DataType exists."""
        assert hasattr(DataType, "VISUAL_JOINT_POSITIONS")
        assert DataType.VISUAL_JOINT_POSITIONS.value == "VISUAL_JOINT_POSITIONS"

    def test_visual_joint_positions_data_type_mapped_to_nc_data_class(self):
        """Test VISUAL_JOINT_POSITIONS mapped to JointData."""
        assert DataType.VISUAL_JOINT_POSITIONS in DATA_TYPE_TO_NC_DATA_CLASS
        assert DATA_TYPE_TO_NC_DATA_CLASS[DataType.VISUAL_JOINT_POSITIONS] == JointData

    def test_visual_joint_positions_data_type_mapped_to_batched_nc_data_class(self):
        """Test VISUAL_JOINT_POSITIONS mapped to BatchedJointData."""
        assert DataType.VISUAL_JOINT_POSITIONS in DATA_TYPE_TO_BATCHED_NC_DATA_CLASS
        assert (
            DATA_TYPE_TO_BATCHED_NC_DATA_CLASS[DataType.VISUAL_JOINT_POSITIONS]
            == BatchedJointData
        )

    def test_visual_joint_positions_uses_same_data_class_as_state(self):
        """Test that visual joint positions and state
        use the same underlying data classes."""
        assert (
            DATA_TYPE_TO_NC_DATA_CLASS[DataType.VISUAL_JOINT_POSITIONS]
            == DATA_TYPE_TO_NC_DATA_CLASS[DataType.JOINT_POSITIONS]
        )
        assert (
            DATA_TYPE_TO_BATCHED_NC_DATA_CLASS[DataType.VISUAL_JOINT_POSITIONS]
            == DATA_TYPE_TO_BATCHED_NC_DATA_CLASS[DataType.JOINT_POSITIONS]
        )


class TestJointTargetPositionsDataType:
    """Tests for JOINT_TARGET_POSITIONS DataType mapping."""

    def test_target_positions_data_type_exists(self):
        """Test that JOINT_TARGET_POSITIONS DataType exists."""
        assert hasattr(DataType, "JOINT_TARGET_POSITIONS")
        assert DataType.JOINT_TARGET_POSITIONS.value == "JOINT_TARGET_POSITIONS"

    def test_target_positions_data_type_mapped_to_nc_data_class(self):
        """Test that JOINT_TARGET_POSITIONS is mapped to JointData."""
        assert DataType.JOINT_TARGET_POSITIONS in DATA_TYPE_TO_NC_DATA_CLASS
        assert DATA_TYPE_TO_NC_DATA_CLASS[DataType.JOINT_TARGET_POSITIONS] == JointData

    def test_target_positions_data_type_mapped_to_batched_nc_data_class(self):
        """Test that JOINT_TARGET_POSITIONS is mapped to BatchedJointData."""
        assert DataType.JOINT_TARGET_POSITIONS in DATA_TYPE_TO_BATCHED_NC_DATA_CLASS
        assert (
            DATA_TYPE_TO_BATCHED_NC_DATA_CLASS[DataType.JOINT_TARGET_POSITIONS]
            == BatchedJointData
        )

    def test_target_uses_same_data_class_as_state(self):
        """Test that target and state use the same underlying data classes."""
        assert (
            DATA_TYPE_TO_NC_DATA_CLASS[DataType.JOINT_TARGET_POSITIONS]
            == DATA_TYPE_TO_NC_DATA_CLASS[DataType.JOINT_POSITIONS]
        )
        assert (
            DATA_TYPE_TO_BATCHED_NC_DATA_CLASS[DataType.JOINT_TARGET_POSITIONS]
            == DATA_TYPE_TO_BATCHED_NC_DATA_CLASS[DataType.JOINT_POSITIONS]
        )


class TestVisualJointPositionsDataImportConfig:
    """Tests for VisualJointPositionsDataImportConfig class."""

    def test_gripper_type_all_indexes_provided(self):
        """Test VisualJointPositionsDataImportConfig with GRIPPER type."""
        config = VisualJointPositionsDataImportConfig(
            source="gripper",
            format=DataFormat(visual_joint_type=VisualJointTypeConfig.GRIPPER),
            mapping=[
                MappingItem(name="gripper_joint_0", index=0),
                MappingItem(name="gripper_joint_1", index=1),
            ],
        )
        assert config.mapping[0].index == 0
        assert config.mapping[1].index == 1

    def test_custom_type_auto_index(self):
        """Test CUSTOM type auto-assigns indexes."""
        config = VisualJointPositionsDataImportConfig(
            source="joints",
            format=DataFormat(visual_joint_type=VisualJointTypeConfig.CUSTOM),
            mapping=[
                MappingItem(name="joint_0"),
                MappingItem(name="joint_1"),
                MappingItem(name="joint_2"),
            ],
        )
        assert config.mapping[0].index == 0
        assert config.mapping[1].index == 1
        assert config.mapping[2].index == 2

    def test_custom_type_all_indexes_provided(self):
        """Test CUSTOM type with all indexes provided."""
        config = VisualJointPositionsDataImportConfig(
            source="joints",
            format=DataFormat(visual_joint_type=VisualJointTypeConfig.CUSTOM),
            mapping=[
                MappingItem(name="joint_0", index=5),
                MappingItem(name="joint_1", index=6),
            ],
        )
        assert config.mapping[0].index == 5
        assert config.mapping[1].index == 6

    def test_custom_type_mixed_indexes_invalid(self):
        """Test validation with CUSTOM type and mixed indexes."""
        with pytest.raises(
            ValueError,
            match="All or none of the mapping items in",
        ):
            VisualJointPositionsDataImportConfig(
                source="joints",
                format=DataFormat(visual_joint_type=VisualJointTypeConfig.CUSTOM),
                mapping=[
                    MappingItem(name="joint_0", index=0),
                    MappingItem(name="joint_1"),  # Missing index
                ],
            )

    def test_gripper_type_transforms(self):
        """Test VisualJointPositionsDataImportConfig transforms for GRIPPER type."""
        config = VisualJointPositionsDataImportConfig(
            source="gripper",
            format=DataFormat(visual_joint_type=VisualJointTypeConfig.GRIPPER),
            mapping=[MappingItem(name="gripper_joint_0", index=0)],
        )
        transforms = config.mapping[0].transforms.transforms
        # Should have Clip, Unnormalize, and NumpyToScalar
        assert any(isinstance(t, Clip) for t in transforms)
        assert any(isinstance(t, Unnormalize) for t in transforms)
        assert isinstance(transforms[-1], NumpyToScalar)

    def test_gripper_type_transforms_with_inverted(self):
        """Test GRIPPER type transforms with inverted flag."""
        config = VisualJointPositionsDataImportConfig(
            source="gripper",
            format=DataFormat(visual_joint_type=VisualJointTypeConfig.GRIPPER),
            mapping=[MappingItem(name="gripper_joint_0", index=0, inverted=True)],
        )
        transforms = config.mapping[0].transforms.transforms
        assert any(isinstance(t, Clip) for t in transforms)
        assert any(isinstance(t, Unnormalize) for t in transforms)
        assert any(isinstance(t, FlipSign) for t in transforms)
        assert isinstance(transforms[-1], NumpyToScalar)

    def test_gripper_type_transforms_with_offset(self):
        """Test GRIPPER type transforms with offset."""
        config = VisualJointPositionsDataImportConfig(
            source="gripper",
            format=DataFormat(visual_joint_type=VisualJointTypeConfig.GRIPPER),
            mapping=[MappingItem(name="gripper_joint_0", index=0, offset=1.5)],
        )
        transforms = config.mapping[0].transforms.transforms
        assert any(isinstance(t, Clip) for t in transforms)
        assert any(isinstance(t, Unnormalize) for t in transforms)
        assert any(isinstance(t, Offset) for t in transforms)
        assert isinstance(transforms[-1], NumpyToScalar)

    def test_custom_type_transforms_degrees_inverted_offset(self):
        """Test CUSTOM type transforms with degrees, inverted, and offset."""
        config = VisualJointPositionsDataImportConfig(
            source="joints",
            format=DataFormat(
                visual_joint_type=VisualJointTypeConfig.CUSTOM,
                angle_units=AngleConfig.DEGREES,
            ),
            mapping=[
                MappingItem(
                    name="joint_0",
                    inverted=True,
                    offset=2.0,
                )
            ],
        )
        transforms = config.mapping[0].transforms.transforms
        assert any(isinstance(t, DegreesToRadians) for t in transforms)
        assert any(isinstance(t, FlipSign) for t in transforms)
        assert any(isinstance(t, Offset) for t in transforms)
        assert isinstance(transforms[-1], NumpyToScalar)

    def test_gripper_type_multiple_joints(self):
        """Test GRIPPER type with multiple joints."""
        config = VisualJointPositionsDataImportConfig(
            source="gripper",
            format=DataFormat(visual_joint_type=VisualJointTypeConfig.GRIPPER),
            mapping=[
                MappingItem(name="gripper_joint_0", index=0),
                MappingItem(name="gripper_joint_1", index=1),
                MappingItem(name="gripper_joint_2", index=2),
            ],
        )
        assert len(config.mapping) == 3
        for i, item in enumerate(config.mapping):
            assert item.index == i
            transforms = item.transforms.transforms
            assert any(isinstance(t, Clip) for t in transforms)
            assert any(isinstance(t, Unnormalize) for t in transforms)
            assert isinstance(transforms[-1], NumpyToScalar)



"""Tests for Custom1DData and BatchedCustom1DData."""

import json

import numpy as np
import torch

from neuracore_types import BatchedCustom1DData, Custom1DData
from neuracore_types.importer.data_config import MappingItem
from neuracore_types.importer.transform import FlipSign, Offset
from neuracore_types.nc_data.custom_1d_data import Custom1DDataImportConfig


class TestCustom1DData:
    """Tests for Custom1DData functionality."""

    def test_sample(self):
        """Test Custom1DData.sample() creates valid instance."""
        data = Custom1DData.sample()
        assert isinstance(data, Custom1DData)
        assert isinstance(data.data, np.ndarray)
        assert data.data.shape == (10,)
        assert data.data.dtype == np.float32
        assert data.type == "Custom1DData"

    def test_serialization(self):
        """Test JSON serialization of custom 1D data."""
        arr = np.array([1.5, 2.5, 3.5, 4.5], dtype=np.float32)
        data = Custom1DData(data=arr)

        json_str = data.model_dump_json()
        loaded = Custom1DData.model_validate_json(json_str)

        assert np.allclose(loaded.data, data.data)

    def test_list_encoding_decoding(self):
        """Test list encoding/decoding."""
        arr = np.random.randn(20).astype(np.float32)
        data = Custom1DData(data=arr)

        # Serialize to dict
        data_dict = json.loads(data.model_dump_json())
        assert isinstance(data_dict["data"], list)

        # Deserialize
        loaded = Custom1DData.model_validate(data_dict)
        assert np.allclose(loaded.data, arr)

    def test_single_element_array(self):
        """Test custom 1D data with single element."""
        arr = np.array([5.0], dtype=np.float32)
        data = Custom1DData(data=arr)

        assert data.data.shape == (1,)
        assert data.data[0] == 5.0

    def test_large_array(self):
        """Test custom 1D data with large array."""
        arr = np.random.randn(10000).astype(np.float32)
        data = Custom1DData(data=arr)

        assert data.data.shape == (10000,)

    def test_empty_array(self):
        """Test handling of empty array."""
        arr = np.array([], dtype=np.float32)
        data = Custom1DData(data=arr)

        assert data.data.shape == (0,)

    def test_all_zeros(self):
        """Test custom 1D data with all zeros."""
        arr = np.zeros(50, dtype=np.float32)
        data = Custom1DData(data=arr)

        assert np.allclose(data.data, 0.0)

    def test_all_ones(self):
        """Test custom 1D data with all ones."""
        arr = np.ones(50, dtype=np.float32)
        data = Custom1DData(data=arr)

        assert np.allclose(data.data, 1.0)


class TestBatchedCustom1DData:
    """Tests for BatchedCustom1DData functionality."""

    def test_from_nc_data(self):
        """Test BatchedCustom1DData.from_nc_data() conversion."""
        custom_data = Custom1DData(data=np.array([1.0, 2.0, 3.0], dtype=np.float32))
        batched = BatchedCustom1DData.from_nc_data(custom_data)

        assert isinstance(batched, BatchedCustom1DData)
        assert batched.data.shape == (1, 1, 3)
        assert torch.allclose(batched.data[0, 0], torch.tensor([1.0, 2.0, 3.0]))

    def test_sample(self):
        """Test BatchedCustom1DData.sample() with different dimensions."""
        batched = BatchedCustom1DData.sample(batch_size=2, time_steps=3)
        assert batched.data.shape == (2, 3, 10)

    def test_sample_single_dimension(self):
        """Test sample with single batch and timestep."""
        batched = BatchedCustom1DData.sample(batch_size=1, time_steps=1)
        assert batched.data.shape == (1, 1, 10)

    def test_sample_large_dimensions(self):
        """Test sample with large dimensions."""
        batched = BatchedCustom1DData.sample(batch_size=50, time_steps=100)
        assert batched.data.shape == (50, 100, 10)

    def test_to_device(self):
        """Test moving to different device."""
        batched = BatchedCustom1DData.sample(batch_size=2, time_steps=3)
        batched_cpu = batched.to(torch.device("cpu"))

        assert batched_cpu.data.device.type == "cpu"
        assert torch.allclose(batched_cpu.data, batched.data)

    def test_from_nc_data_preserves_values(self):
        """Test that from_nc_data preserves exact values."""
        arr = np.array([1.5, 2.5, 3.5, 4.5, 5.5], dtype=np.float32)
        custom_data = Custom1DData(data=arr)
        batched = BatchedCustom1DData.from_nc_data(custom_data)

        assert torch.allclose(batched.data[0, 0], torch.from_numpy(arr), rtol=1e-5)

    def test_can_serialize_deserialize(self):
        """Test JSON serialization and deserialization of BatchedCustom1DData."""
        batched = BatchedCustom1DData.sample(batch_size=2, time_steps=2)
        json_str = batched.model_dump_json()
        loaded = BatchedCustom1DData.model_validate_json(json_str)

        assert torch.equal(loaded.data, batched.data)
        assert loaded.data.shape == batched.data.shape


class TestCustom1DDataSerialization:
    """Tests for serialization of Custom1DData."""

    def test_roundtrip_serialization(self):
        """Test that data survives roundtrip serialization."""
        original_arr = np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)
        data = Custom1DData(data=original_arr)

        # Serialize
        json_str = data.model_dump_json()

        # Deserialize
        loaded = Custom1DData.model_validate_json(json_str)

        # Verify
        assert np.allclose(loaded.data, original_arr)

    def test_serialization_preserves_dtype(self):
        """Test that serialization preserves float32 dtype."""
        arr = np.array([1.0, 2.0, 3.0], dtype=np.float32)
        data = Custom1DData(data=arr)

        json_str = data.model_dump_json()
        loaded = Custom1DData.model_validate_json(json_str)

        assert loaded.data.dtype == np.float32

    def test_base64_is_reversible(self):
        """Test that base64 encoding is fully reversible."""
        arr = np.random.randn(100).astype(np.float32)
        data = Custom1DData(data=arr)

        # Encode
        encoded = data.model_dump()

        # Decode
        loaded = Custom1DData.model_validate(encoded)

        # Should be exactly equal (not just close)
        assert np.array_equal(loaded.data, arr)


class TestCustom1DDataImportConfig:
    """Tests for Custom1DDataImportConfig class."""

    def test_custom_1d_data_import_config_basic(self):
        """Test Custom1DDataImportConfig basic functionality."""
        data_point = Custom1DDataImportConfig(
            source="custom",
            mapping=[MappingItem(name="custom_value")],
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert len(transforms) == 0

    def test_custom_1d_data_import_config_inverted(self):
        """Test Custom1DDataImportConfig with inverted flag."""
        data_point = Custom1DDataImportConfig(
            source="custom",
            mapping=[MappingItem(name="custom_value", inverted=True)],
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert any(isinstance(t, FlipSign) for t in transforms)

    def test_custom_1d_data_import_config_offset(self):
        """Test Custom1DDataImportConfig with offset."""
        data_point = Custom1DDataImportConfig(
            source="custom",
            mapping=[MappingItem(name="custom_value", offset=2.5)],
        )
        transforms = data_point.mapping[0].transforms.transforms
        assert any(isinstance(t, Offset) for t in transforms)



"""Tests for data ordering functionality.

This module contains tests to verify that the data ordering utilities
work correctly and ensure consistent ordering across sync points.
"""

import time

import numpy as np

from neuracore_types import (
    Custom1DData,
    DataType,
    DepthCameraData,
    JointData,
    LanguageData,
    PointCloudData,
    PoseData,
    RGBCameraData,
    SynchronizedEpisode,
    SynchronizedPoint,
)

SYNCHRONIZED_POINT_UNORDERED = SynchronizedPoint(
    timestamp=time.time(),
    data={
        DataType.JOINT_POSITIONS: {
            "joint_2": JointData(value=0.1),
            "joint_1": JointData(value=0.2),
            "joint_3": JointData(value=0.3),
        },
        DataType.JOINT_VELOCITIES: {
            "joint_2": JointData(value=0.1),
            "joint_1": JointData(value=0.2),
            "joint_3": JointData(value=0.3),
        },
        DataType.POSES: {
            "pose_2": PoseData(pose=[0.1, 0.2, 0.3, 0.0, 0.0, 0.0, 1.0]),
            "pose_1": PoseData(pose=[0.1, 0.2, 0.3, 0.0, 0.0, 0.0, 1.0]),
        },
        DataType.RGB_IMAGES: {
            "camera_3": RGBCameraData(frame=np.zeros((224, 224, 3))),
            "camera_1": RGBCameraData(frame=np.zeros((224, 224, 3))),
            "camera_2": RGBCameraData(frame=np.zeros((224, 224, 3))),
        },
        DataType.DEPTH_IMAGES: {
            "depth_2": DepthCameraData(frame=np.zeros((224, 224))),
            "depth_1": DepthCameraData(frame=np.zeros((224, 224))),
        },
        DataType.POINT_CLOUDS: {
            "lidar_2": PointCloudData(
                points=np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float16)
            ),
            "lidar_1": PointCloudData(points=np.array([[7, 8, 9]], dtype=np.float16)),
        },
        DataType.CUSTOM_1D: {
            "sensor_z": Custom1DData(data=np.array([1, 2, 3])),
            "sensor_a": Custom1DData(data=np.array([4, 5, 6])),
        },
        DataType.LANGUAGE: {"default": LanguageData(text="Pick up the object")},
    },
)

ORDER_SCHEMA = {
    DataType.JOINT_POSITIONS: sorted(
        SYNCHRONIZED_POINT_UNORDERED.data[DataType.JOINT_POSITIONS].keys()
    ),
    DataType.JOINT_VELOCITIES: sorted(
        SYNCHRONIZED_POINT_UNORDERED.data[DataType.JOINT_VELOCITIES].keys()
    ),
    DataType.POSES: sorted(SYNCHRONIZED_POINT_UNORDERED.data[DataType.POSES].keys()),
    DataType.RGB_IMAGES: sorted(
        SYNCHRONIZED_POINT_UNORDERED.data[DataType.RGB_IMAGES].keys()
    ),
    DataType.DEPTH_IMAGES: sorted(
        SYNCHRONIZED_POINT_UNORDERED.data[DataType.DEPTH_IMAGES].keys()
    ),
    DataType.POINT_CLOUDS: sorted(
        SYNCHRONIZED_POINT_UNORDERED.data[DataType.POINT_CLOUDS].keys()
    ),
    DataType.LANGUAGE: sorted(
        SYNCHRONIZED_POINT_UNORDERED.data[DataType.LANGUAGE].keys()
    ),
    DataType.CUSTOM_1D: sorted(
        SYNCHRONIZED_POINT_UNORDERED.data[DataType.CUSTOM_1D].keys()
    ),
}


def _verify_sync_point_ordering(sync_point_unordered: SynchronizedPoint):
    """Test that sync point ordering works correctly."""
    ordered_sync = sync_point_unordered.order(ORDER_SCHEMA)

    # Verify that all keys are now sorted
    assert DataType.JOINT_POSITIONS in ordered_sync.data
    assert list(ordered_sync[DataType.JOINT_POSITIONS].keys()) == [
        "joint_1",
        "joint_2",
        "joint_3",
    ]

    # Verify joint velocities ordering
    assert DataType.JOINT_VELOCITIES in ordered_sync.data
    assert list(ordered_sync[DataType.JOINT_VELOCITIES].keys()) == [
        "joint_1",
        "joint_2",
        "joint_3",
    ]

    # Verify poses ordering (both the dict keys and pose names should be sorted)
    assert DataType.POSES in ordered_sync.data
    assert list(ordered_sync[DataType.POSES].keys()) == ["pose_1", "pose_2"]

    # Verify RGB images ordering
    assert DataType.RGB_IMAGES in ordered_sync.data
    assert list(ordered_sync[DataType.RGB_IMAGES].keys()) == [
        "camera_1",
        "camera_2",
        "camera_3",
    ]

    # Verify depth images ordering
    assert DataType.DEPTH_IMAGES in ordered_sync.data
    assert list(ordered_sync[DataType.DEPTH_IMAGES].keys()) == ["depth_1", "depth_2"]

    # Verify point clouds ordering
    assert DataType.POINT_CLOUDS in ordered_sync.data
    assert list(ordered_sync[DataType.POINT_CLOUDS].keys()) == ["lidar_1", "lidar_2"]

    # Verify custom data ordering
    assert DataType.CUSTOM_1D in ordered_sync.data
    assert list(ordered_sync[DataType.CUSTOM_1D].keys()) == ["sensor_a", "sensor_z"]


def test_sync_point_ordering():
    """Test that sync point ordering works correctly."""
    _verify_sync_point_ordering(SYNCHRONIZED_POINT_UNORDERED)


def test_synced_data_ordering():
    """Test that synced data ordering works for multiple sync points."""
    # Create multiple unordered sync points
    sync_points = [SYNCHRONIZED_POINT_UNORDERED for _ in range(3)]

    # Modify timestamps to be different
    for i, sync_point in enumerate(sync_points):
        sync_point.timestamp = time.time() + i

    # Create synced data
    synced_data = SynchronizedEpisode(
        observations=sync_points,
        start_time=sync_points[0].timestamp,
        end_time=sync_points[-1].timestamp,
        robot_id="robot1",
    )

    ordered_synced_data = synced_data.order(ORDER_SCHEMA)

    for frame in ordered_synced_data.observations:
        _verify_sync_point_ordering(frame)



"""Pose data types for 6DOF poses."""

import copy
from typing import Literal, Union

import numpy as np
from pydantic import (
    ConfigDict,
    Field,
    field_serializer,
    field_validator,
    model_validator,
)

from neuracore_types.importer.config import AngleConfig, PoseConfig, RotationConfig
from neuracore_types.importer.transform import (
    DataTransform,
    DataTransformSequence,
    Pose,
)
from neuracore_types.nc_data.nc_data import (
    DataItemStats,
    NCData,
    NCDataImportConfig,
    NCDataStats,
)
from neuracore_types.utils.pydantic_to_ts import (
    REQUIRED_WITH_DEFAULT_FLAG,
    fix_required_with_defaults,
)


class PoseDataStats(NCDataStats):
    """Statistics for PoseData."""

    type: Literal["PoseDataStats"] = Field(
        default="PoseDataStats", json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    pose: DataItemStats

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)


class PoseDataImportConfig(NCDataImportConfig):
    """Import configuration for PoseData."""

    @model_validator(mode="after")
    def validate_orientation_required(self) -> "PoseDataImportConfig":
        """Validate orientation is provided when format is position_orientation."""
        if self.format.pose_type == PoseConfig.POSITION_ORIENTATION:
            if self.format.orientation is None:
                raise ValueError(
                    "orientation must be provided when format is 'position_orientation'"
                )
            if self.format.orientation.type == RotationConfig.QUATERNION:
                if not self.format.orientation.quaternion_order:
                    raise ValueError(
                        "quaternion_order must be provided when type is 'quaternion'"
                    )
            if self.format.orientation.type == RotationConfig.EULER:
                if not self.format.orientation.euler_order:
                    raise ValueError(
                        "euler_order must be provided when type is 'euler'"
                    )
        return self

    @model_validator(mode="after")
    def validate_index_range(self) -> "PoseDataImportConfig":
        """Validate that index range length matches the format."""
        if self.format.pose_type == PoseConfig.MATRIX:
            return self

        for item in self.mapping:
            if item.index_range is None:
                raise ValueError("index_range is required for pose data points")
            index_length = item.index_range.end - item.index_range.start

            if self.format.pose_type == PoseConfig.MATRIX:
                if index_length != 16:
                    raise ValueError(
                        "Index range length must be 16 for matrix format, "
                        f"got {index_length}"
                    )
            elif self.format.pose_type == PoseConfig.POSITION_ORIENTATION:
                if self.format.orientation is None:
                    raise ValueError(
                        "orientation is required when pose_type is "
                        "'position_orientation'"
                    )
                if self.format.orientation.type == RotationConfig.QUATERNION:
                    expected_length = 7  # 3 position + 4 quaternion
                elif self.format.orientation.type in [
                    RotationConfig.EULER,
                    RotationConfig.AXIS_ANGLE,
                ]:  # euler or axis_angle
                    expected_length = 6  # 3 position + 3 euler or axis_angle
                elif self.format.orientation.type == RotationConfig.MATRIX:
                    expected_length = 9  # 3 position + 3x3 matrix
                else:
                    raise ValueError(
                        f"Unsupported orientation type: {self.format.orientation.type}"
                    )
                if index_length != expected_length:
                    raise ValueError(
                        f"Index range length must be {expected_length} for "
                        f"orientation type {self.format.orientation.type}, "
                        f"got {index_length}"
                    )

        return self

    def _populate_transforms(self) -> None:
        """Populate transforms based on flags if not already set."""
        transform_list: list[DataTransform] = []
        # Add Pose transform based on format and orientation settings
        for item in self.mapping:
            item_transforms = copy.deepcopy(transform_list)
            if self.format.pose_type == PoseConfig.MATRIX:
                item_transforms.append(Pose(pose_type=PoseConfig.MATRIX))
            elif self.format.pose_type == PoseConfig.POSITION_ORIENTATION:
                if self.format.orientation is None:
                    raise ValueError(
                        "orientation is required when pose_type is "
                        "'position_orientation'"
                    )
                # Determine sequence
                seq: str = "xyzw"
                if self.format.orientation.type == RotationConfig.QUATERNION:
                    seq = self.format.orientation.quaternion_order.value
                elif self.format.orientation.type == RotationConfig.EULER:
                    seq = self.format.orientation.euler_order.value

                item_transforms.append(
                    Pose(
                        pose_type=PoseConfig.POSITION_ORIENTATION,
                        rotation_type=RotationConfig(self.format.orientation.type),
                        angle_type=AngleConfig(self.format.orientation.angle_units),
                        seq=seq,
                    )
                )
            item.transforms = DataTransformSequence(transforms=item_transforms)


class PoseData(NCData):
    """6DOF pose data for objects, end-effectors, or coordinate frames.

    Represents position and orientation information for tracking objects
    or robot components in 3D space. Poses are stored as dictionaries
    mapping pose names to [x, y, z, rx, ry, rz] values.
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True, json_schema_extra=fix_required_with_defaults
    )

    type: Literal["PoseData"] = Field(
        default="PoseData", json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    pose: np.ndarray

    @field_validator("pose")
    @classmethod
    def validate_pose_length(cls, v: np.ndarray) -> np.ndarray:
        """Validate that pose has exactly 7 values (position + quaternion)."""
        if len(v) != 7:
            raise ValueError(
                "Pose must have exactly 7 values "
                f"(x, y, z, qx, qy, qz, qw), got {len(v)}"
            )
        return v

    @field_validator("pose", mode="before")
    @classmethod
    def decode_pose(cls, v: Union[list, np.ndarray]) -> np.ndarray:
        """Decode pose to NumPy array."""
        return np.array(v, dtype=np.float32) if isinstance(v, list) else v

    @field_serializer("pose", when_used="json")
    def serialize_pose(self, v: np.ndarray) -> list[float]:
        """Serialize pose to JSON list."""
        return v.tolist()

    @classmethod
    def sample(cls) -> "PoseData":
        """Sample an example PoseData instance.

        Returns:
            PoseData: Sampled PoseData instance
        """
        return cls(pose=np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0]))

    def calculate_statistics(self) -> PoseDataStats:
        """Calculate the statistics for this data type.

        Returns:
            Dictionary attribute names to their corresponding DataItemStats.
        """
        stats = DataItemStats(
            mean=np.array(self.pose, dtype=np.float32),
            std=np.zeros_like(np.array(self.pose, dtype=np.float32)),
            count=np.array([1] * len(self.pose), dtype=np.int32),
            min=np.array(self.pose, dtype=np.float32),
            max=np.array(self.pose, dtype=np.float32),
        )
        return PoseDataStats(pose=stats)



"""Neuracore Types - Shared type definitions for Neuracore."""

from neuracore_types.dataset import *  # noqa: F403
from neuracore_types.endpoints import *  # noqa: F403
from neuracore_types.episode import *  # noqa: F403
from neuracore_types.nc_data import *  # noqa: F403
from neuracore_types.synchronization import *  # noqa: F403
from neuracore_types.training import *  # noqa: F403
from neuracore_types.upload import *  # noqa: F403

TORCH_AVAILABLE = True
try:
    import torch  # noqa: F401

    del torch
except ImportError:
    TORCH_AVAILABLE = False

if TORCH_AVAILABLE:
    from neuracore_types.batched_nc_data import *  # noqa: F403

__version__ = "6.8.0"



"""Data models for parallel gripper open amount data."""

import copy
from typing import Literal

import numpy as np
from pydantic import ConfigDict, Field

from neuracore_types.importer.transform import (
    Clip,
    DataTransform,
    DataTransformSequence,
    FlipSign,
    Normalize,
    NumpyToScalar,
    Offset,
)
from neuracore_types.nc_data.nc_data import (
    DataItemStats,
    NCData,
    NCDataImportConfig,
    NCDataStats,
)
from neuracore_types.utils.pydantic_to_ts import (
    REQUIRED_WITH_DEFAULT_FLAG,
    fix_required_with_defaults,
)


class ParallelGripperOpenAmountDataStats(NCDataStats):
    """Statistics for ParallelGripperOpenAmountData."""

    type: Literal["ParallelGripperOpenAmountDataStats"] = Field(
        default="ParallelGripperOpenAmountDataStats",
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )
    open_amount: DataItemStats

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)


class ParallelGripperOpenAmountDataImportConfig(NCDataImportConfig):
    """Import configuration for ParallelGripperOpenAmountData."""

    def _populate_transforms(self) -> None:
        """Populate transforms based on configuration."""
        transform_list: list[DataTransform] = []

        # Add Normalize transform if needed
        if self.format.normalize:
            transform_list.append(
                Normalize(min=self.format.normalize.min, max=self.format.normalize.max)
            )

        # Clip the value to 0-1
        transform_list.append(Clip(min=0.0, max=1.0))

        # Convert from close amount to open amount if needed
        if self.format.invert_gripper_amount:
            transform_list.append(FlipSign())
            transform_list.append(Offset(value=1.0))

        for item in self.mapping:
            item_transforms = copy.deepcopy(transform_list)
            if item.inverted:
                item_transforms.append(FlipSign())
            if item.offset != 0.0:
                item_transforms.append(Offset(value=item.offset))
            item_transforms.append(NumpyToScalar())
            item.transforms = DataTransformSequence(transforms=item_transforms)


class ParallelGripperOpenAmountData(NCData):
    """Open amount data for parallel end effector gripper.

    Contains the state of parallel gripper opening amounts.
    """

    type: Literal["ParallelGripperOpenAmountData"] = Field(
        default="ParallelGripperOpenAmountData",
        json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG,
    )
    open_amount: float

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)

    @classmethod
    def sample(cls) -> "ParallelGripperOpenAmountData":
        """Sample an example ParallelGripperOpenAmountData instance.

        Returns:
            ParallelGripperOpenAmountData: Sampled instance
        """
        return cls(open_amount=0.0)

    def calculate_statistics(self) -> ParallelGripperOpenAmountDataStats:
        """Calculate the statistics for this data type.

        Returns:
            Dictionary attribute names to their corresponding DataItemStats.
        """
        stats = DataItemStats(
            mean=np.array([self.open_amount], dtype=np.float32),
            std=np.array([0.0], dtype=np.float32),
            count=np.array([1], dtype=np.int32),
            min=np.array([self.open_amount], dtype=np.float32),
            max=np.array([self.open_amount], dtype=np.float32),
        )
        return ParallelGripperOpenAmountDataStats(open_amount=stats)



"""Request models for dataset and recording synchronization operations."""

from pydantic import BaseModel, ConfigDict, Field

from neuracore_types.episode.episode import RobotDataSpec
from neuracore_types.utils.pydantic_to_ts import (
    REQUIRED_WITH_DEFAULT_FLAG,
    fix_required_with_defaults,
)


class SynchronizationDetails(BaseModel):
    """Details for synchronization requests.

    Attributes:
        frequency: Synchronization frequency in Hz.
        robot_data_spec: Specification of robot data to include in the synchronization.
        max_delay_s: Maximum allowable delay (in seconds) for synchronization.
        allow_duplicates: Whether to allow duplicate data points in the synchronization.
        trim_start_end: Whether to trim the start and end of the episode
            when synchronizing.
    """

    frequency: int
    robot_data_spec: RobotDataSpec | None
    max_delay_s: float = Field(
        default=0.1, json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    allow_duplicates: bool = Field(
        default=True, json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    trim_start_end: bool = Field(
        default=True, json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )

    model_config = ConfigDict(frozen=True, json_schema_extra=fix_required_with_defaults)

    def __hash__(self) -> int:
        """Compute a hash value for the SynchronizationDetails instance.

        Returns:
            int: The computed hash value.
        """
        # Convert the nested dict structure to something hashable
        robot_data_spec_hashable = None
        if self.robot_data_spec is not None:
            # Convert dict[str, dict[DataType, list[str]]] to a frozen structure
            robot_data_spec_hashable = tuple(
                sorted(
                    (
                        robot_name,
                        tuple(
                            sorted(
                                (data_type, tuple(fields))
                                for data_type, fields in data_spec.items()
                            )
                        ),
                    )
                    for robot_name, data_spec in self.robot_data_spec.items()
                )
            )

        return hash((
            self.frequency,
            robot_data_spec_hashable,
            self.max_delay_s,
            self.allow_duplicates,
            self.trim_start_end,
        ))



"""Joint data types for robot joint states."""

import copy
from typing import Literal

import numpy as np
from pydantic import ConfigDict, Field, model_validator

from neuracore_types.importer.config import (
    AngleConfig,
    JointPositionTypeConfig,
    PoseConfig,
    RotationConfig,
    TorqueUnitsConfig,
    VisualJointTypeConfig,
)
from neuracore_types.importer.data_config import MappingItem
from neuracore_types.importer.transform import (
    Clip,
    DataTransform,
    DataTransformSequence,
    DegreesToRadians,
    FlipSign,
    Normalize,
    NumpyToScalar,
    Offset,
    Pose,
    Scale,
    Unnormalize,
)
from neuracore_types.nc_data.nc_data import (
    DataItemStats,
    NCData,
    NCDataImportConfig,
    NCDataStats,
)
from neuracore_types.utils.pydantic_to_ts import (
    REQUIRED_WITH_DEFAULT_FLAG,
    fix_required_with_defaults,
)


def _validate_index_provided(mapping: list[MappingItem], class_name: str) -> None:
    """Validate that either all or no indexes are provided for mapping items."""
    indexes = [item.index for item in mapping]
    if any(idx is not None for idx in indexes) and any(idx is None for idx in indexes):
        raise ValueError(
            f"All or none of the mapping items in {class_name} must have an "
            "'index' specified"
        )
    # If no indexes are provided, assign them sequentially
    if not any(idx is not None for idx in indexes):
        for i, item in enumerate(mapping):
            item.index = i


def _apply_common_joint_item_transforms(
    item: MappingItem, transforms_list: list[DataTransform]
) -> list[DataTransform]:
    """Apply common transforms for joint data items (flip and offset)."""
    item_transforms = copy.deepcopy(transforms_list)
    if item.inverted:
        item_transforms.append(FlipSign())
    if getattr(item, "offset", 0.0) != 0.0:
        item_transforms.append(Offset(value=item.offset))
    item_transforms.append(NumpyToScalar())
    return item_transforms


class JointDataStats(NCDataStats):
    """Statistics for JointData."""

    type: Literal["JointDataStats"] = Field(
        default="JointDataStats", json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    value: DataItemStats

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)


class JointPositionsDataImportConfig(NCDataImportConfig):
    """Import configuration for JointPositionsData."""

    @model_validator(mode="after")
    def validate_orientation_required(self) -> "JointPositionsDataImportConfig":
        """Validate orientation is when converting joint position type from TCP."""
        if self.format.joint_position_type == JointPositionTypeConfig.END_EFFECTOR:
            if self.format.pose_type == PoseConfig.POSITION_ORIENTATION:
                if self.format.orientation is None:
                    raise ValueError(
                        "orientation must be provided when format is "
                        "'position_orientation'"
                    )
                if self.format.orientation.type == RotationConfig.QUATERNION:
                    if not self.format.orientation.quaternion_order:
                        raise ValueError(
                            "quaternion_order must be provided when type is "
                            "'quaternion'"
                        )
                if self.format.orientation.type == RotationConfig.EULER:
                    if not self.format.orientation.euler_order:
                        raise ValueError(
                            "euler_order must be provided when type is 'euler'"
                        )
        return self

    @model_validator(mode="after")
    def validate_index_provided(self) -> "JointPositionsDataImportConfig":
        """Validate that indexes provided are valid.

        When directly importing joint positions, check either all or no indexes
        are provided. When converting from TCP to joint positions, check that
        the index range length matches the orientation format.
        """
        if self.format.joint_position_type == JointPositionTypeConfig.CUSTOM:
            _validate_index_provided(self.mapping, self.__class__.__name__)
        elif self.format.joint_position_type == JointPositionTypeConfig.END_EFFECTOR:
            if len(self.mapping) != 1:
                raise ValueError(
                    "Only one mapping item is allowed when converting from TCP "
                    "to joint positions"
                )

            if self.format.pose_type == PoseConfig.MATRIX:
                return self
            else:
                if self.mapping[0].index_range is None:
                    raise ValueError("index_range is required for pose data points")
                index_length = (
                    self.mapping[0].index_range.end - self.mapping[0].index_range.start
                )

                if self.format.pose_type == PoseConfig.MATRIX:
                    if index_length != 16:
                        raise ValueError(
                            "Index range length must be 16 for matrix format, "
                            f"got {index_length}"
                        )
                elif self.format.pose_type == PoseConfig.POSITION_ORIENTATION:
                    if self.format.orientation is None:
                        raise ValueError(
                            "orientation is required when pose_type is "
                            "'position_orientation'"
                        )
                    if self.format.orientation.type == RotationConfig.QUATERNION:
                        expected_length = 7  # 3 position + 4 quaternion
                    elif self.format.orientation.type in [
                        RotationConfig.EULER,
                        RotationConfig.AXIS_ANGLE,
                    ]:  # euler or axis_angle
                        expected_length = 6  # 3 position + 3 euler or axis_angle
                    elif self.format.orientation.type == RotationConfig.MATRIX:
                        expected_length = 9  # 3 position + 3x3 matrix
                    else:
                        raise ValueError(
                            f"Unsupported orientation type: "
                            f"{self.format.orientation.type}"
                        )
                    if index_length != expected_length:
                        raise ValueError(
                            f"Index range length must be {expected_length} for "
                            f"orientation type {self.format.orientation.type}, "
                            f"got {index_length}"
                        )
        return self

    def _populate_transforms(self) -> None:
        """Populate transforms based on configuration."""
        transform_list: list[DataTransform] = []

        if self.format.joint_position_type == JointPositionTypeConfig.CUSTOM:
            # Add DegreesToRadians transform if needed
            if self.format.angle_units == AngleConfig.DEGREES:
                transform_list.append(DegreesToRadians())

            for item in self.mapping:
                item_transforms = _apply_common_joint_item_transforms(
                    item, transform_list
                )
                item.transforms = DataTransformSequence(transforms=item_transforms)
        elif self.format.joint_position_type == JointPositionTypeConfig.END_EFFECTOR:
            for item in self.mapping:
                item_transforms = copy.deepcopy(transform_list)
                if self.format.pose_type == PoseConfig.MATRIX:
                    item_transforms.append(Pose(pose_type=PoseConfig.MATRIX))
                elif self.format.pose_type == PoseConfig.POSITION_ORIENTATION:
                    if self.format.orientation is None:
                        raise ValueError(
                            "orientation is required when pose_type is "
                            "'position_orientation'"
                        )
                    # Determine sequence
                    seq: str = "xyzw"
                    if self.format.orientation.type == RotationConfig.QUATERNION:
                        seq = self.format.orientation.quaternion_order.value
                    elif self.format.orientation.type == RotationConfig.EULER:
                        seq = self.format.orientation.euler_order.value

                    item_transforms.append(
                        Pose(
                            pose_type=PoseConfig.POSITION_ORIENTATION,
                            rotation_type=RotationConfig(self.format.orientation.type),
                            angle_type=AngleConfig(self.format.orientation.angle_units),
                            seq=seq,
                        )
                    )
                item.transforms = DataTransformSequence(transforms=item_transforms)
        else:
            raise ValueError(
                f"Invalid joint position type: {self.format.joint_position_type}"
            )


class JointVelocitiesDataImportConfig(NCDataImportConfig):
    """Import configuration for JointVelocitiesData."""

    @model_validator(mode="after")
    def validate_index_provided(self) -> "JointVelocitiesDataImportConfig":
        """Validate that either all or no indexes are provided."""
        _validate_index_provided(self.mapping, self.__class__.__name__)
        return self

    def _populate_transforms(self) -> None:
        """Populate transforms based on configuration."""
        transform_list: list[DataTransform] = []

        if self.format.angle_units == AngleConfig.DEGREES:
            transform_list.append(DegreesToRadians())

        for item in self.mapping:
            item_transforms = _apply_common_joint_item_transforms(item, transform_list)
            item.transforms = DataTransformSequence(transforms=item_transforms)


class JointTorquesDataImportConfig(NCDataImportConfig):
    """Import configuration for JointTorquesData."""

    @model_validator(mode="after")
    def validate_index_provided(self) -> "JointTorquesDataImportConfig":
        """Validate that either all or no indexes are provided."""
        _validate_index_provided(self.mapping, self.__class__.__name__)
        return self

    def _populate_transforms(self) -> None:
        """Populate transforms based on configuration."""
        transform_list: list[DataTransform] = []

        if self.format.torque_units == TorqueUnitsConfig.NCM:
            transform_list.append(Scale(factor=0.01))

        for item in self.mapping:
            item_transforms = _apply_common_joint_item_transforms(item, transform_list)
            item.transforms = DataTransformSequence(transforms=item_transforms)


class VisualJointPositionsDataImportConfig(NCDataImportConfig):
    """Import configuration for VisualJointPositionsData."""

    @model_validator(mode="after")
    def validate_index_provided(self) -> "VisualJointPositionsDataImportConfig":
        """Validate that either all or no indexes are provided."""
        if self.format.visual_joint_type == VisualJointTypeConfig.CUSTOM:
            _validate_index_provided(self.mapping, self.__class__.__name__)
        elif self.format.visual_joint_type != VisualJointTypeConfig.GRIPPER:
            raise ValueError(
                f"Invalid visual joint type: {self.format.visual_joint_type}"
            )

        return self

    def _populate_transforms(self) -> None:
        """Populate transforms based on configuration."""
        transform_list: list[DataTransform] = []

        if self.format.visual_joint_type == VisualJointTypeConfig.GRIPPER:
            # Use gripper open amount to calculate the visual joint positions
            # from joint limits

            # Add Normalize transform if needed
            if self.format.normalize:
                transform_list.append(
                    Normalize(
                        min=self.format.normalize.min, max=self.format.normalize.max
                    )
                )

            # Clip the value to 0-1
            transform_list.append(Clip(min=0.0, max=1.0))

            # Convert from close amount to open amount if needed
            if self.format.invert_gripper_amount:
                transform_list.append(FlipSign())
                transform_list.append(Offset(value=1.0))

            # Joint limits to be filled in during data import from URDF
            transform_list.append(Unnormalize(min=0.0, max=1.0))

        elif self.format.visual_joint_type == VisualJointTypeConfig.CUSTOM:
            # Add DegreesToRadians transform if needed
            if self.format.angle_units == AngleConfig.DEGREES:
                transform_list.append(DegreesToRadians())
        else:
            raise ValueError(
                f"Invalid visual joint type: {self.format.visual_joint_type}"
            )

        for item in self.mapping:
            item_transforms = _apply_common_joint_item_transforms(item, transform_list)
            item.transforms = DataTransformSequence(transforms=item_transforms)


class JointData(NCData):
    """Robot joint state data including positions, velocities, or torques."""

    type: Literal["JointData"] = Field(
        default="JointData", json_schema_extra=REQUIRED_WITH_DEFAULT_FLAG
    )
    value: float

    model_config = ConfigDict(json_schema_extra=fix_required_with_defaults)

    def calculate_statistics(self) -> JointDataStats:
        """Calculate the statistics for this data type.

        Returns:
            Dictionary attribute names to their corresponding DataItemStats.
        """
        stats = DataItemStats(
            mean=np.array([self.value], dtype=np.float32),
            std=np.array([0.0], dtype=np.float32),
            count=np.array([1], dtype=np.int32),
            min=np.array([self.value], dtype=np.float32),
            max=np.array([self.value], dtype=np.float32),
        )
        return JointDataStats(value=stats)

    @classmethod
    def sample(cls) -> "JointData":
        """Sample an example JointData instance.

        Returns:
            JointData: Sampled JointData instance
        """
        return cls(value=0.0)
